# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import re

import numpy as np
import pytest
from pydantic.v1 import BaseModel

from braket.circuits import AngledGate, FreeParameter, FreeParameterExpression, Gate
from braket.circuits.angled_gate import DoubleAngledGate, TripleAngledGate


@pytest.fixture
def angled_gate():
    return AngledGate(angle=0.15, qubit_count=1, ascii_symbols=["foo"])


def test_is_operator(angled_gate):
    assert isinstance(angled_gate, Gate)


def test_angle_is_none():
    with pytest.raises(ValueError, match="angle must not be None"):
        AngledGate(qubit_count=1, ascii_symbols=["foo"], angle=None)


def test_getters():
    qubit_count = 2
    ascii_symbols = ("foo", "bar")
    angle = 0.15
    gate = AngledGate(angle=angle, qubit_count=qubit_count, ascii_symbols=ascii_symbols)

    assert gate.qubit_count == qubit_count
    assert gate.ascii_symbols == ascii_symbols
    assert gate.angle == angle


def test_angle_setter(angled_gate):
    with pytest.raises(AttributeError):
        angled_gate.angle = 0.14


def test_adjoint(angled_gate):
    assert angled_gate.adjoint() == [AngledGate(angle=-0.15, qubit_count=1, ascii_symbols=["foo"])]


def test_adjoint_parameterized():
    theta = FreeParameter("theta")
    angled_gate = AngledGate(angle=(theta + 1), qubit_count=1, ascii_symbols=["foo"])
    assert angled_gate.adjoint() == [
        AngledGate(angle=-(theta + 1), qubit_count=1, ascii_symbols=["foo"])
    ]


def test_equality(angled_gate):
    gate = AngledGate(angle=0.15, qubit_count=1, ascii_symbols=["bar"])
    other_gate = AngledGate(angle=0.3, qubit_count=1, ascii_symbols=["foo"])
    non_gate = "non gate"

    assert angled_gate == gate
    assert angled_gate is not gate
    assert angled_gate != other_gate
    assert angled_gate != non_gate


def test_symbolic_equality():
    symbol1 = FreeParameter("theta")
    symbol2 = FreeParameter("phi")
    symbol3 = FreeParameter("theta")
    gate1 = AngledGate(angle=symbol1, qubit_count=1, ascii_symbols=["bar"])
    gate2 = AngledGate(angle=symbol1, qubit_count=1, ascii_symbols=["bar"])
    gate3 = AngledGate(angle=symbol3, qubit_count=1, ascii_symbols=["bar"])
    other_gate = AngledGate(angle=symbol2, qubit_count=1, ascii_symbols=["foo"])

    assert gate1 == gate2
    assert gate1 == gate3
    assert gate1 is not gate2
    assert gate1 != other_gate


def test_mixed_angle_equality():
    symbol1 = FreeParameter("theta")
    gate1 = AngledGate(angle=symbol1, qubit_count=1, ascii_symbols=["bar"])
    gate2 = AngledGate(angle=0.15, qubit_count=1, ascii_symbols=["foo"])

    assert gate1 != gate2
    assert gate2 != gate1


def test_angle_adjoint():
    symbol1 = FreeParameter("theta")
    gate1 = AngledGate(angle=symbol1, qubit_count=1, ascii_symbols=["bar(theta)"])
    gate2 = AngledGate(angle=0.15, qubit_count=1, ascii_symbols=["foo(0.15)"])

    gate1_adj = gate1.adjoint()
    gate2_adj = gate2.adjoint()

    assert len(gate1_adj) == len(gate2_adj) == 1
    assert np.array_equal(gate1_adj[0].ascii_symbols, ["bar(-theta)"])
    assert np.array_equal(gate2_adj[0].ascii_symbols, ["foo(-0.15)"])


def test_bind_values():
    theta = FreeParameter("theta")
    gate = AngledGate(angle=theta, qubit_count=1, ascii_symbols=["bar"])
    with pytest.raises(NotImplementedError):
        gate.bind_values(theta=1)


def test_angled_gate_with_expr():
    expr = FreeParameterExpression(FreeParameter("theta") + 1)
    new_expr = expr.subs({"theta": 1})
    gate = AngledGate(angle=new_expr, qubit_count=1, ascii_symbols=["bar"])
    expected = AngledGate(angle=2, qubit_count=1, ascii_symbols=["bar"])

    assert gate == expected


def test_np_float_angle_json():
    angled_gate = AngledGate(angle=np.float32(0.15), qubit_count=1, ascii_symbols=["foo"])
    angled_gate_json = BaseModel.construct(target=[0], angle=angled_gate.angle).json()
    match = re.match(r'\{"target": \[0], "angle": (\d*\.?\d*)}', angled_gate_json)
    angle_value = float(match.group(1))
    assert angle_value == angled_gate.angle


def test_double_angle_is_none():
    with pytest.raises(ValueError, match="angles must not be None"):
        DoubleAngledGate(qubit_count=1, ascii_symbols=["foo"], angle_1=None, angle_2=1)


def test_triple_angle_is_none():
    with pytest.raises(ValueError, match="angles must not be None"):
        TripleAngledGate(qubit_count=1, ascii_symbols=["foo"], angle_1=None, angle_2=1, angle_3=2)


def test_double_angle_equality():
    gate = DoubleAngledGate(angle_1=0.15, angle_2=3, qubit_count=1, ascii_symbols=["bar"])
    equal_gate = DoubleAngledGate(angle_1=0.15, angle_2=3, qubit_count=1, ascii_symbols=["bar"])
    other_gate = AngledGate(angle=0.3, qubit_count=1, ascii_symbols=["foo"])
    non_gate = "non gate"

    assert equal_gate == gate
    assert equal_gate is not gate
    assert gate != other_gate
    assert gate != non_gate


def test_double_angle_symbolic_equality():
    symbol1 = FreeParameter("theta")
    symbol2 = FreeParameter("phi")
    symbol3 = FreeParameter("theta")
    gate1 = DoubleAngledGate(angle_1=symbol1, angle_2=1, qubit_count=1, ascii_symbols=["bar"])
    gate2 = DoubleAngledGate(angle_1=symbol1, angle_2=1, qubit_count=1, ascii_symbols=["bar"])
    gate3 = DoubleAngledGate(angle_1=symbol3, angle_2=1, qubit_count=1, ascii_symbols=["bar"])
    other_gate = DoubleAngledGate(angle_1=symbol2, angle_2=1, qubit_count=1, ascii_symbols=["foo"])

    assert gate1 == gate2
    assert gate1 == gate3
    assert gate1 is not gate2
    assert gate1 != other_gate


def test_double_angle_repr():
    assert (
        repr(DoubleAngledGate(qubit_count=1, ascii_symbols=["foo"], angle_1=1, angle_2=2))
        == "DoubleAngledGate('angles': (1.0, 2.0), 'qubit_count': 1)"
    )


def test_triple_angle_repr():
    assert (
        repr(
            TripleAngledGate(qubit_count=1, ascii_symbols=["foo"], angle_1=1, angle_2=2, angle_3=3)
        )
        == "TripleAngledGate('angles': (1.0, 2.0, 3.0), 'qubit_count': 1)"
    )


def test_double_angle_parameters():
    assert DoubleAngledGate(
        qubit_count=1, ascii_symbols=["foo"], angle_1=1, angle_2=2
    ).parameters == [1, 2]


def test_hash_double_angle():
    symbol1 = FreeParameter("theta")
    assert hash(
        DoubleAngledGate(angle_1=symbol1, angle_2=1, qubit_count=1, ascii_symbols=["bar"])
    ) == hash(DoubleAngledGate(angle_1=symbol1, angle_2=1, qubit_count=1, ascii_symbols=["bar"]))


def test_hash_triple_angle():
    symbol1 = FreeParameter("theta")
    assert hash(
        TripleAngledGate(
            angle_1=symbol1, angle_2=1, angle_3=3, qubit_count=1, ascii_symbols=["bar"]
        )
    ) == hash(
        TripleAngledGate(
            angle_1=symbol1, angle_2=1, angle_3=3, qubit_count=1, ascii_symbols=["bar"]
        )
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import functools
import importlib.util
import inspect
import re
import shutil
import sys
import tempfile
import warnings
from collections.abc import Callable, Iterable
from logging import Logger, getLogger
from pathlib import Path
from types import ModuleType
from typing import Any

import cloudpickle

from braket.aws.aws_session import AwsSession
from braket.jobs._entry_point_template import run_entry_point, symlink_input_data
from braket.jobs.config import (
    CheckpointConfig,
    InstanceConfig,
    OutputDataConfig,
    S3DataSourceConfig,
    StoppingCondition,
)
from braket.jobs.image_uris import Framework, built_in_images, retrieve_image
from braket.jobs.quantum_job import QuantumJob
from braket.jobs.quantum_job_creation import _generate_default_job_name


def hybrid_job(
    *,
    device: str | None,
    include_modules: str | ModuleType | Iterable[str | ModuleType] | None = None,
    dependencies: str | Path | list[str] | None = None,
    local: bool = False,
    job_name: str | None = None,
    image_uri: str | None = None,
    input_data: str | dict | S3DataSourceConfig | None = None,
    wait_until_complete: bool = False,
    instance_config: InstanceConfig | None = None,
    distribution: str | None = None,
    copy_checkpoints_from_job: str | None = None,
    checkpoint_config: CheckpointConfig | None = None,
    role_arn: str | None = None,
    stopping_condition: StoppingCondition | None = None,
    output_data_config: OutputDataConfig | None = None,
    aws_session: AwsSession | None = None,
    tags: dict[str, str] | None = None,
    logger: Logger = getLogger(__name__),
    quiet: bool | None = None,
    reservation_arn: str | None = None,
) -> Callable:
    """Defines a hybrid job by decorating the entry point function. The job will be created
    when the decorated function is called.

    The job created will be a `LocalQuantumJob` when `local` is set to `True`, otherwise an
    `AwsQuantumJob`. The following parameters will be ignored when running a job with
    `local` set to `True`: `wait_until_complete`, `instance_config`, `distribution`,
    `copy_checkpoints_from_job`, `stopping_condition`, `tags`, `logger`, and `quiet`.

    Args:
        device (str | None): Device ARN of the QPU device that receives priority quantum
            task queueing once the hybrid job begins running. Each QPU has a separate hybrid jobs
            queue so that only one hybrid job is running at a time. The device string is accessible
            in the hybrid job instance as the environment variable "AMZN_BRAKET_DEVICE_ARN".
            When using embedded simulators, you may provide the device argument as string of the
            form: "local:<provider>/<simulator_name>" or `None`.

        include_modules (str | ModuleType | Iterable[str | ModuleType] | None): Either a
            single module or module name or a list of module or module names referring to local
            modules to be included. Any references to members of these modules in the hybrid job
            algorithm code will be serialized as part of the algorithm code. Default: `[]`

        dependencies (str | Path | list[str] | None): Path (absolute or relative) to a
            requirements.txt file, or alternatively a list of strings, with each string being a
            `requirement specifier <https://pip.pypa.io/en/stable/reference/requirement-specifiers/
            #requirement-specifiers>`_, to be used for the hybrid job.

        local (bool): Whether to use local mode for the hybrid job. Default: `False`

        job_name (str | None): A string that specifies the name with which the job is created.
            Allowed pattern for job name: `^[a-zA-Z0-9](-*[a-zA-Z0-9]){0,50}$`. Defaults to
            f'{decorated-function-name}-{timestamp}'.

        image_uri (str | None): A str that specifies the ECR image to use for executing the job.
            `retrieve_image()` function may be used for retrieving the ECR image URIs
            for the containers supported by Braket. Default: `<Braket base image_uri>`.

        input_data (str | dict | S3DataSourceConfig | None): Information about the training
            data. Dictionary maps channel names to local paths or S3 URIs. Contents found
            at any local paths will be uploaded to S3 at
            f's3://{default_bucket_name}/jobs/{job_name}/data/{channel_name}'. If a local
            path, S3 URI, or S3DataSourceConfig is provided, it will be given a default
            channel name "input".
            Default: {}.

        wait_until_complete (bool): `True` if we should wait until the job completes.
            This would tail the job logs as it waits. Otherwise `False`. Ignored if using
            local mode. Default: `False`.

        instance_config (InstanceConfig | None): Configuration of the instance(s) for running the
            classical code for the hybrid job. Default:
            `InstanceConfig(instanceType='ml.m5.large', instanceCount=1, volumeSizeInGB=30)`.

        distribution (str | None): A str that specifies how the job should be distributed.
            If set to "data_parallel", the hyperparameters for the job will be set to use data
            parallelism features for PyTorch or TensorFlow. Default: `None`.

        copy_checkpoints_from_job (str | None): A str that specifies the job ARN whose
            checkpoint you want to use in the current job. Specifying this value will copy
            over the checkpoint data from `use_checkpoints_from_job`'s checkpoint_config
            s3Uri to the current job's checkpoint_config s3Uri, making it available at
            checkpoint_config.localPath during the job execution. Default: `None`

        checkpoint_config (CheckpointConfig | None): Configuration that specifies the
            location where checkpoint data is stored.
            Default: `CheckpointConfig(localPath='/opt/jobs/checkpoints',
            s3Uri=f's3://{default_bucket_name}/jobs/{job_name}/checkpoints')`.

        role_arn (str | None): A str providing the IAM role ARN used to execute the
            script. Default: IAM role returned by AwsSession's `get_default_jobs_role()`.

        stopping_condition (StoppingCondition | None): The maximum length of time, in seconds,
            and the maximum number of tasks that a job can run before being forcefully stopped.
            Default: StoppingCondition(maxRuntimeInSeconds=5 * 24 * 60 * 60).

        output_data_config (OutputDataConfig | None): Specifies the location for the output of
            the job.
            Default: `OutputDataConfig(s3Path=f's3://{default_bucket_name}/jobs/{job_name}/data',
            kmsKeyId=None)`.

        aws_session (AwsSession | None): AwsSession for connecting to AWS Services.
            Default: AwsSession()

        tags (dict[str, str] | None): Dict specifying the key-value pairs for tagging this job.
            Default: {}.

        logger (Logger): Logger object with which to write logs, such as task statuses
            while waiting for task to be in a terminal state. Default: `getLogger(__name__)`

        quiet (bool | None): Sets the verbosity of the logger to low and does not report queue
            position. Default is `False`.

        reservation_arn (str | None): the reservation window arn provided by Braket
            Direct to reserve exclusive usage for the device to run the hybrid job on.
            Default: None.

    Returns:
        Callable: the callable for creating a Hybrid Job.
    """
    _validate_python_version(image_uri, aws_session)

    def _hybrid_job(entry_point: Callable) -> Callable:
        @functools.wraps(entry_point)
        def job_wrapper(*args: Any, **kwargs: Any) -> Callable:
            """The job wrapper.

            Args:
                *args (Any): Arbitrary arguments.
                **kwargs (Any): Arbitrary keyword arguments.

            Returns:
                Callable: the callable for creating a Hybrid Job.
            """
            with _IncludeModules(include_modules), tempfile.TemporaryDirectory(
                dir="", prefix="decorator_job_"
            ) as temp_dir:
                temp_dir_path = Path(temp_dir)
                entry_point_file_path = Path("entry_point.py")
                with open(temp_dir_path / entry_point_file_path, "w") as entry_point_file:
                    template = "\n".join(
                        [
                            _process_input_data(input_data),
                            _serialize_entry_point(entry_point, args, kwargs),
                        ]
                    )
                    entry_point_file.write(template)

                if dependencies:
                    _process_dependencies(dependencies, temp_dir_path)

                job_args = {
                    "device": device or "local:none/none",
                    "source_module": temp_dir,
                    "entry_point": (
                        f"{temp_dir}.{entry_point_file_path.stem}:{entry_point.__name__}"
                    ),
                    "wait_until_complete": wait_until_complete,
                    "job_name": job_name or _generate_default_job_name(func=entry_point),
                    "hyperparameters": _log_hyperparameters(entry_point, args, kwargs),
                    "logger": logger,
                }
                optional_args = {
                    "image_uri": image_uri,
                    "input_data": input_data,
                    "instance_config": instance_config,
                    "distribution": distribution,
                    "checkpoint_config": checkpoint_config,
                    "copy_checkpoints_from_job": copy_checkpoints_from_job,
                    "role_arn": role_arn,
                    "stopping_condition": stopping_condition,
                    "output_data_config": output_data_config,
                    "aws_session": aws_session,
                    "tags": tags,
                    "quiet": quiet,
                    "reservation_arn": reservation_arn,
                }
                for key, value in optional_args.items():
                    if value is not None:
                        job_args[key] = value

                job = _create_job(job_args, local)
            return job

        return job_wrapper

    return _hybrid_job


def _validate_python_version(image_uri: str | None, aws_session: AwsSession | None = None) -> None:
    """Validate python version at job definition time"""
    aws_session = aws_session or AwsSession()
    # user provides a custom image_uri
    if image_uri and image_uri not in built_in_images(aws_session.region):
        print(
            "Skipping python version validation, make sure versions match "
            "between local environment and container."
        )
    else:
        # set default image_uri to base
        image_uri = image_uri or retrieve_image(Framework.BASE, aws_session.region)
        tag = aws_session.get_full_image_tag(image_uri)
        major_version, minor_version = re.search(r"-py(\d)(\d+)-", tag).groups()
        if not (sys.version_info.major, sys.version_info.minor) == (
            int(major_version),
            int(minor_version),
        ):
            raise RuntimeError(
                "Python version must match between local environment and container. "
                f"Client is running Python {sys.version_info.major}.{sys.version_info.minor} "
                f"locally, but container uses Python {major_version}.{minor_version}."
            )


def _process_dependencies(dependencies: str | Path | list[str], temp_dir: Path) -> None:
    if isinstance(dependencies, (str, Path)):
        # requirements file
        shutil.copy(Path(dependencies).resolve(), temp_dir / "requirements.txt")
    else:
        # list of packages
        with open(temp_dir / "requirements.txt", "w") as f:
            f.write("\n".join(dependencies))


class _IncludeModules:
    def __init__(self, modules: str | ModuleType | Iterable[str | ModuleType] = None):
        modules = modules or []
        if isinstance(modules, (str, ModuleType)):
            modules = [modules]
        self._modules = [
            (importlib.import_module(module) if isinstance(module, str) else module)
            for module in modules
        ]

    def __enter__(self):
        """Register included modules with cloudpickle to be pickled by value"""
        for module in self._modules:
            cloudpickle.register_pickle_by_value(module)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Unregister included modules with cloudpickle to be pickled by value"""
        for module in self._modules:
            cloudpickle.unregister_pickle_by_value(module)


def _serialize_entry_point(entry_point: Callable, args: tuple, kwargs: dict) -> str:
    """Create an entry point from a function"""
    wrapped_entry_point = functools.partial(entry_point, *args, **kwargs)

    try:
        serialized = cloudpickle.dumps(wrapped_entry_point)
    except Exception as e:
        raise RuntimeError(
            "Serialization failed for decorator hybrid job. If you are referencing "
            "an object from outside the function scope, either directly or through "
            "function parameters, try instantiating the object inside the decorated "
            "function instead."
        ) from e

    return run_entry_point.format(
        serialized=serialized,
        function_name=entry_point.__name__,
    )


def _log_hyperparameters(entry_point: Callable, args: tuple, kwargs: dict) -> dict:
    """Capture function arguments as hyperparameters"""
    signature = inspect.signature(entry_point)
    bound_args = signature.bind(*args, **kwargs)
    bound_args.apply_defaults()
    hyperparameters = {}
    for param, value in bound_args.arguments.items():
        param_kind = signature.parameters[param].kind
        if param_kind in [
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.KEYWORD_ONLY,
        ]:
            hyperparameters[param] = value
        elif param_kind == inspect.Parameter.VAR_KEYWORD:
            hyperparameters.update(**value)
        else:
            warnings.warn(
                "Positional only arguments will not be logged to the hyperparameters file.",
                stacklevel=1,
            )
    return {name: _sanitize(value) for name, value in hyperparameters.items()}


def _sanitize(hyperparameter: Any) -> str:
    """Sanitize forbidden characters from hp strings"""
    string_hp = str(hyperparameter)

    sanitized = (
        string_hp
        # replace forbidden characters with close matches
        .replace("\n", " ")
        .replace("$", "?")
        .replace("(", "{")
        .replace("&", "+")
        .replace("`", "'")
        # not technically forbidden, but to avoid mismatched parens
        .replace(")", "}")
    )

    # max allowed length for a hyperparameter is 2500
    if len(sanitized) > 2500:
        # show as much as possible, including the final 20 characters
        return f"{sanitized[:2500 - 23]}...{sanitized[-20:]}"
    return sanitized


def _process_input_data(input_data: dict) -> list[str]:
    """Create symlinks to data.

    Logic chart for how the service moves files into the data directory on the instance:
        input data matches exactly one file: cwd/filename -> channel/filename
        input data matches exactly one directory: cwd/dirname/* -> channel/*
        else (multiple matches, possibly including exact):
            cwd/prefix_match -> channel/prefix_match, for each match
    """
    input_data = input_data or {}
    if not isinstance(input_data, dict):
        input_data = {"input": input_data}

    def matches(prefix: str) -> list[str]:
        return [
            str(path) for path in Path(prefix).parent.iterdir() if str(path).startswith(str(prefix))
        ]

    def is_prefix(path: str) -> bool:
        return len(matches(path)) > 1 or not Path(path).exists()

    prefix_channels = set()
    directory_channels = set()
    file_channels = set()

    for channel, data in input_data.items():
        if AwsSession.is_s3_uri(str(data)) or isinstance(data, S3DataSourceConfig):
            channel_arg = f'channel="{channel}"' if channel != "input" else ""
            print(
                "Input data channels mapped to an S3 source will not be available in "
                f"the working directory. Use `get_input_data_dir({channel_arg})` to read "
                f"input data from S3 source inside the job container."
            )
        elif is_prefix(data):
            prefix_channels.add(channel)
        elif Path(data).is_dir():
            directory_channels.add(channel)
        else:
            file_channels.add(channel)

    return symlink_input_data.format(
        prefix_matches={channel: matches(input_data[channel]) for channel in prefix_channels},
        input_data_items=[
            (channel, data)
            for channel, data in input_data.items()
            if channel in prefix_channels | directory_channels | file_channels
        ],
        prefix_channels=prefix_channels,
        directory_channels=directory_channels,
    )


def _create_job(job_args: dict[str, Any], local: bool = False) -> QuantumJob:
    """Create an AWS or Local hybrid job"""
    if local:
        from braket.jobs.local import LocalQuantumJob

        for aws_only_arg in [
            "wait_until_complete",
            "copy_checkpoints_from_job",
            "instance_config",
            "distribution",
            "stopping_condition",
            "tags",
            "logger",
        ]:
            if aws_only_arg in job_args:
                del job_args[aws_only_arg]
        return LocalQuantumJob.create(**job_args)
    else:
        from braket.aws import AwsQuantumJob

        return AwsQuantumJob.create(**job_args)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from braket.circuits import Circuit, ResultType


def validate_circuit_and_shots(circuit: Circuit, shots: int) -> None:
    """Validates if circuit and shots are correct before running on a device

    Args:
        circuit (Circuit): circuit to validate
        shots (int): shots to validate

    Raises:
        ValueError: If circuit has no instructions; if circuit has a non-gphase instruction; if no
            result types specified for circuit and `shots=0`. See `braket.circuit.result_types`;
            if circuit has observables that cannot be simultaneously measured and `shots>0`;
            or, if `StateVector` or `Amplitude` are specified as result types when `shots>0`.
    """
    if not circuit.instructions or all(
        not (inst.target or inst.control) for inst in circuit.instructions
    ):
        raise ValueError("Circuit must have at least one non-zero-qubit gate to run on a device")
    if not shots and not circuit.result_types:
        raise ValueError(
            "No result types specified for circuit and shots=0. See `braket.circuits.result_types`"
        )
    elif shots and circuit.result_types:
        if not circuit.observables_simultaneously_measurable:
            raise ValueError("Observables cannot be sampled simultaneously")
        for rt in circuit.result_types:
            if isinstance(rt, (ResultType.Amplitude, ResultType.StateVector)):
                raise ValueError("StateVector or Amplitude cannot be specified when shots>0")
            elif isinstance(rt, ResultType.Probability):
                num_qubits = len(rt.target) or circuit.qubit_count
                if num_qubits > 40:
                    raise ValueError("Probability target must be less than or equal to 40 qubits.")



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

import braket.ir.jaqcd as ir
from braket.circuits import compiler_directives
from braket.circuits.compiler_directive import CompilerDirective
from braket.circuits.serialization import IRType

testdata = [
    (
        compiler_directives.StartVerbatimBox,
        ir.StartVerbatimBox,
        "#pragma braket verbatim\nbox{",
        compiler_directives.EndVerbatimBox,
    ),
    (
        compiler_directives.EndVerbatimBox,
        ir.EndVerbatimBox,
        "}",
        compiler_directives.StartVerbatimBox,
    ),
]


@pytest.mark.parametrize("testclass,irclass,openqasm_str,counterpart", testdata)
def test_counterpart(testclass, irclass, openqasm_str, counterpart):
    assert testclass().counterpart() == counterpart()


@pytest.mark.parametrize("testclass,irclass,openqasm_str,counterpart", testdata)
def test_to_ir(testclass, irclass, openqasm_str, counterpart):
    assert testclass().to_ir(ir_type=IRType.JAQCD) == irclass()
    assert testclass().to_ir(ir_type=IRType.OPENQASM) == openqasm_str


@pytest.mark.parametrize("testclass,irclass,openqasm_str,counterpart", testdata)
def test_equality(testclass, irclass, openqasm_str, counterpart):
    op1 = testclass()
    op2 = testclass()
    assert op1 == op2
    assert op1 is not op2
    assert op1 != CompilerDirective(ascii_symbols=["foo"])
    assert op1 != "not a directive"



import importlib
import multiprocessing
import os
import pathlib

import pytest


def test_for_import_cycles():
    # Note, because of all the multiprocessing in this test, when running 'tox', the process
    # threads may be made to wait as other tests are running in parallel, making it seems like
    # this test is much slower than it actually is. However, splitting the test into a
    # parameterized version wasn't able to correctly detect some circular imports when running tox.
    modules = get_modules_to_test()
    processes = []
    multiprocessing.set_start_method("spawn", force=True)
    for module in modules:
        # We create a separate process to make sure the imports do not interfere with each-other.
        process = multiprocessing.Process(target=import_module, args=(module,))
        processes.append(process)
        process.start()

    for index, process in enumerate(processes):
        process.join()
        if process.exitcode != 0:
            pytest.fail(
                f"Unable to import '{modules[index]}'."
                " If all other tests are passing, check for cyclical dependencies."
            )


def get_modules_to_test():
    curr_path = pathlib.Path(__file__).resolve()
    while "test" in str(curr_path):
        curr_path = curr_path.parent
    curr_path = curr_path.joinpath("src")
    curr_path_len = len(str(curr_path)) + len(os.sep)
    modules = []
    for dir_, temp, files in os.walk(curr_path):
        # Rather than testing every single python file we just test modules, for now.
        if "__init__.py" in files:
            braket_module = dir_[curr_path_len:].replace(os.sep, ".")
            modules.append(braket_module)
    return modules


def import_module(module):
    importlib.import_module(module)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import random
import uuid
from unittest.mock import Mock, PropertyMock, patch

import pytest
from common_test_utils import MockS3

from braket.aws import AwsQuantumTaskBatch, AwsSession
from braket.circuits import Circuit
from braket.tasks import GateModelQuantumTaskResult

S3_TARGET = AwsSession.S3DestinationFolder("foo", "bar")


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_creation(mock_create):
    task_mock = Mock()
    type(task_mock).id = PropertyMock(side_effect=uuid.uuid4)
    task_mock.state.return_value = "RUNNING"
    mock_create.return_value = task_mock

    batch_size = 10
    batch = AwsQuantumTaskBatch(
        Mock(),
        "foo",
        _circuits(batch_size),
        S3_TARGET,
        1000,
        max_parallel=10,
        reservaion_arn=(
            "arn:aws:braket:us-west-2:123456789123:"
            "reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9"
        ),
    )
    assert batch.size == batch_size
    assert batch.tasks == [task_mock for _ in range(batch_size)]
    assert len(batch.unfinished) == batch_size
    assert not batch.unsuccessful


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_successful(mock_create):
    task_mock = Mock()
    type(task_mock).id = PropertyMock(side_effect=uuid.uuid4)
    task_mock.state.return_value = "COMPLETED"
    result = GateModelQuantumTaskResult.from_string(MockS3.MOCK_S3_RESULT_GATE_MODEL)
    task_mock.result.return_value = result
    mock_create.return_value = task_mock

    batch_size = 15
    batch = AwsQuantumTaskBatch(
        Mock(),
        "foo",
        _circuits(batch_size),
        S3_TARGET,
        1000,
        max_parallel=10,
        reservaion_arn=(
            "arn:aws:braket:us-west-2:123456789123:"
            "reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9"
        ),
    )
    assert batch.size == batch_size
    assert not batch.unfinished
    assert not batch.unsuccessful
    assert batch.results() == [result for _ in range(batch_size)]


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_unsuccessful(mock_create):
    task_mock = Mock()
    task_id = uuid.uuid4()
    type(task_mock).id = PropertyMock(return_value=task_id)
    task_mock.state.return_value = random.choice(["CANCELLED", "FAILED"])
    task_mock.result.return_value = None
    mock_create.return_value = task_mock

    batch = AwsQuantumTaskBatch(
        Mock(),
        "foo",
        [Circuit().h(0).cnot(0, 1)],
        S3_TARGET,
        1000,
        max_parallel=10,
        reservaion_arn=(
            "arn:aws:braket:us-west-2:123456789123:"
            "reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9"
        ),
    )
    assert not batch.unfinished
    assert batch.unsuccessful == {task_id}
    assert batch.results() == [None]
    with pytest.raises(RuntimeError):
        assert batch.results(fail_unsuccessful=True) == [None]
    batch._unsuccessful = set()
    with pytest.raises(RuntimeError):
        batch.results(fail_unsuccessful=True, use_cached_value=False)
    assert batch.unsuccessful == {task_id}


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_retry(mock_create):
    bad_task_mock = Mock()
    type(bad_task_mock).id = PropertyMock(side_effect=uuid.uuid4)
    bad_task_mock.state.return_value = random.choice(["CANCELLED", "FAILED"])
    bad_task_mock.result.return_value = None

    good_task_mock = Mock()
    # task id already mocked when setting up bad_task_mock
    good_task_mock.state.return_value = "COMPLETED"
    result = GateModelQuantumTaskResult.from_string(MockS3.MOCK_S3_RESULT_GATE_MODEL)
    good_task_mock.result.return_value = result

    mock_create.side_effect = [bad_task_mock, good_task_mock, bad_task_mock, good_task_mock]

    batch = AwsQuantumTaskBatch(
        Mock(),
        "foo",
        [Circuit().h(0).cnot(0, 1), Circuit().h(1).cnot(0, 1)],
        S3_TARGET,
        1000,
        max_parallel=10,
        reservaion_arn=(
            "arn:aws:braket:us-west-2:123456789123:"
            "reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9"
        ),
    )
    assert not batch.unfinished
    assert batch.results(max_retries=0) == [None, result]

    # Retrying should get rid of the failures
    assert batch.results(fail_unsuccessful=True, max_retries=3, use_cached_value=False) == [
        result,
        result,
    ]
    assert batch.unsuccessful == set()

    # Don't retry if there's nothing to retry
    mock_create.side_effect = [bad_task_mock]
    assert batch.retry_unsuccessful_tasks()
    assert batch.unsuccessful == set()

    # Error if called before there are any results
    batch._results = None
    with pytest.raises(RuntimeError):
        batch.retry_unsuccessful_tasks()


@patch("braket.aws.aws_quantum_task_batch.ThreadPoolExecutor")
def test_abort(mock_threadpool):
    batch_size = 10
    num_workers = 2
    mock_threadpool().__exit__.side_effect = KeyboardInterrupt()

    with pytest.raises(KeyboardInterrupt):
        AwsQuantumTaskBatch(
            Mock(),
            "foo",
            _circuits(batch_size),
            S3_TARGET,
            1000,
            max_parallel=num_workers,
            reservaion_arn=(
                "arn:aws:braket:us-west-2:123456789123:"
                "reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9"
            ),
        )


@patch("concurrent.futures.ThreadPoolExecutor.submit")
def test_early_abort(mock_submit):
    batch_size = 10
    num_workers = 2
    mock_submit.side_effect = [Mock(), KeyboardInterrupt()]

    with pytest.raises(KeyboardInterrupt):
        AwsQuantumTaskBatch(
            Mock(),
            "foo",
            _circuits(batch_size),
            S3_TARGET,
            1000,
            max_parallel=num_workers,
            reservaion_arn=(
                "arn:aws:braket:us-west-2:123456789123:"
                "reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9"
            ),
        )


def _circuits(batch_size):
    return [Circuit().h(0).cnot(0, 1) for _ in range(batch_size)]



import math

import numpy as np
import pytest

from braket.aws import AwsDevice, AwsQuantumTask
from braket.circuits import Circuit, Gate, GateCalibrations, QubitSet
from braket.parametric import FreeParameter
from braket.pulse import ArbitraryWaveform, Frame, Port, PulseSequence


@pytest.fixture
def device():
    return AwsDevice("arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3")


@pytest.fixture
def arbitrary_waveform():
    return ArbitraryWaveform(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.00017888439538396808,
            0.00046751103636033026,
            0.0011372942989106456,
            0.002577059611929697,
            0.005443941944632366,
            0.010731922770068104,
            0.01976701723583167,
            0.03406712171899736,
            0.05503285980691202,
            0.08350670755829034,
            0.11932853352131022,
            0.16107456696238298,
            0.20614055551722368,
            0.2512065440720643,
            0.292952577513137,
            0.328774403476157,
            0.3572482512275353,
            0.3782139893154499,
            0.3925140937986156,
            0.40154918826437913,
            0.4068371690898149,
            0.4097040514225177,
            0.41114381673553674,
            0.411813599998087,
            0.4121022266390633,
            0.4122174383870584,
            0.41226003881132406,
            0.4122746298554775,
            0.4122792591252675,
            0.4122806196003006,
            0.41228098995582513,
            0.41228108334474756,
            0.4122811051578895,
            0.4122811098772742,
            0.4122811108230642,
            0.4122811109986316,
            0.41228111102881937,
            0.41228111103362725,
            0.4122811110343365,
            0.41228111103443343,
            0.4122811110344457,
            0.4122811110344471,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.41228111103444737,
            0.4122811110344471,
            0.4122811110344457,
            0.41228111103443343,
            0.4122811110343365,
            0.41228111103362725,
            0.41228111102881937,
            0.4122811109986316,
            0.4122811108230642,
            0.4122811098772742,
            0.4122811051578895,
            0.41228108334474756,
            0.41228098995582513,
            0.4122806196003006,
            0.4122792591252675,
            0.4122746298554775,
            0.41226003881132406,
            0.4122174383870584,
            0.4121022266390633,
            0.411813599998087,
            0.41114381673553674,
            0.4097040514225176,
            0.4068371690898149,
            0.40154918826437913,
            0.3925140937986155,
            0.37821398931544986,
            0.3572482512275351,
            0.32877440347615655,
            0.2929525775131368,
            0.2512065440720641,
            0.20614055551722307,
            0.16107456696238268,
            0.11932853352131002,
            0.08350670755829034,
            0.05503285980691184,
            0.03406712171899729,
            0.01976701723583167,
            0.010731922770068058,
            0.005443941944632366,
            0.002577059611929697,
            0.0011372942989106229,
            0.00046751103636033026,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )


@pytest.fixture
def port():
    return Port("test_port_ff", dt=1e-9)


@pytest.fixture
def frame_id(device):
    return next(iter(device.frames))


@pytest.fixture
def frame(frame_id, port):
    return Frame(frame_id, port, 1e6, is_predefined=True)


@pytest.fixture
def pulse_sequence(frame, arbitrary_waveform):
    return (
        PulseSequence()
        .barrier(qubits_or_frames=[frame])
        .play(frame=frame, waveform=arbitrary_waveform)
    )


def h_gate(q0):
    return Circuit().rz(q0, np.pi).rx(q0, np.pi / 2).rz(q0, np.pi / 2).rx(q0, -np.pi / 2)


def cz_pulse(
    q0: str,
    q1: str,
    shift_phases_q0: float,
    shift_phases_q1: float,
    waveform: ArbitraryWaveform,
    device: AwsDevice,
):
    q0_rf_frame = device.frames[f"q{q0}_rf_frame"]
    q1_rf_frame = device.frames[f"q{q1}_rf_frame"]
    q0_q1_cz_frame = device.frames[f"q{q0}_q{q1}_cz_frame"]
    frames = [q0_rf_frame, q1_rf_frame, q0_q1_cz_frame]

    dt = device.properties.pulse.ports[q0_q1_cz_frame.port.id].dt
    wfm_duration = len(waveform.amplitudes) * dt

    pulse_sequence = (
        PulseSequence()
        .barrier(frames)
        .play(q0_q1_cz_frame, waveform)
        .delay(q0_rf_frame, wfm_duration)
        .shift_phase(q0_rf_frame, shift_phases_q0)
        .delay(q1_rf_frame, wfm_duration)
        .shift_phase(q1_rf_frame, shift_phases_q1)
        .barrier(frames)
    )
    for phase, q in [(shift_phases_q0 * 0.5, q0), (-shift_phases_q1 * 0.5, q1)]:
        for neighbor in device.properties.paradigm.connectivity.connectivityGraph[str(q)]:
            xy_frame_name = f"q{min(q, int(neighbor))}_q{max(q, int(neighbor))}_xy_frame"
            if xy_frame_name in device.frames:
                xy_frame = device.frames[xy_frame_name]
                pulse_sequence.shift_phase(xy_frame, phase)
    return pulse_sequence


def test_pulse_bell(arbitrary_waveform, device):
    if device.status == "OFFLINE":
        pytest.skip("Device offline")
    (
        a,
        b,
    ) = (
        10,
        113,
    )  # qubits used
    p0, p1 = 1.1733407221086924, 6.269846678712192
    theta_0, theta_1 = FreeParameter("theta_0"), FreeParameter("theta_1")
    a_b_cz_waveform = arbitrary_waveform
    cz = cz_pulse(a, b, theta_0, theta_1, a_b_cz_waveform, device)

    bell_pair_with_gates = Circuit().h(a).h(b).cz(a, b).h(b)
    bell_pair_with_pulses_unbound = (
        h_gate(a) + h_gate(b) + Circuit().pulse_gate([a, b], cz) + h_gate(b)
    )
    bell_pair_with_pulses = bell_pair_with_pulses_unbound(theta_0=p0, theta_1=p1)

    num_shots = 1000
    gate_task = device.run(bell_pair_with_gates, shots=num_shots, disable_qubit_rewiring=True)
    pulse_task = device.run(bell_pair_with_pulses, shots=num_shots)

    if not device.is_available:
        try:
            assert gate_task.state not in AwsQuantumTask.TERMINAL_STATES
            assert pulse_task.state not in AwsQuantumTask.TERMINAL_STATES
        finally:
            gate_task.cancel()
            pulse_task.cancel()
        return

    gate_measurements = gate_task.result().measurement_counts
    pulse_measurements = pulse_task.result().measurement_counts

    # 1-smoothing to avoid nans for 0 counts
    observed = (
        np.array([pulse_measurements.get(state, 1) for state in ("00", "01", "10", "11")])
        / num_shots
    )
    expected = (
        np.array([gate_measurements.get(state, 1) for state in ("00", "01", "10", "11")])
        / num_shots
    )
    chi_squared = np.sum((observed - expected) ** 2 / expected)
    assert chi_squared < 10  # adjust this threshold if test is flaky


def test_pulse_sequence(arbitrary_waveform, device):
    if device.status == "OFFLINE":
        pytest.skip("Device offline")
    (
        a,
        b,
    ) = (
        10,
        113,
    )  # qubits used
    p0, p1 = 1.1733407221086924, 6.269846678712192
    theta_0, theta_1 = FreeParameter("theta_0"), FreeParameter("theta_1")
    a_b_cz_waveform = arbitrary_waveform

    cz_with_pulses_unbound = cz_pulse(a, b, theta_0, theta_1, a_b_cz_waveform, device)

    q0_readout_frame = device.frames[f"q{a}_ro_rx_frame"]
    q1_readout_frame = device.frames[f"q{b}_ro_rx_frame"]
    cz_with_pulses = (
        cz_with_pulses_unbound(theta_0=p0, theta_1=p1)
        .capture_v0(q0_readout_frame)
        .capture_v0(q1_readout_frame)
    )
    cz_with_gates = Circuit().cz(a, b)

    num_shots = 1000
    gate_task = device.run(cz_with_gates, shots=num_shots, disable_qubit_rewiring=True)
    pulse_task = device.run(cz_with_pulses, shots=num_shots)

    if not device.is_available:
        try:
            assert gate_task.state not in AwsQuantumTask.TERMINAL_STATES
            assert pulse_task.state not in AwsQuantumTask.TERMINAL_STATES
        finally:
            gate_task.cancel()
            pulse_task.cancel()
        return

    gate_measurements = gate_task.result().measurement_counts
    pulse_measurements = pulse_task.result().measurement_counts

    # 1-smoothing to avoid nans for 0 counts
    observed = (
        np.array([pulse_measurements.get(state, 1) for state in ("00", "01", "10", "11")])
        / num_shots
    )
    expected = (
        np.array([gate_measurements.get(state, 1) for state in ("00", "01", "10", "11")])
        / num_shots
    )
    chi_squared = np.sum((observed - expected) ** 2 / expected)
    assert chi_squared < 10  # adjust this threshold if test is flaky


def test_gate_calibration_run(device, pulse_sequence):
    if device.status == "OFFLINE":
        pytest.skip("Device offline")
    user_gate_calibrations = GateCalibrations({(Gate.Rx(math.pi / 2), QubitSet(0)): pulse_sequence})
    num_shots = 50
    bell_circuit = Circuit().rx(0, math.pi / 2).rx(1, math.pi / 2).cz(0, 1).rx(1, -math.pi / 2)
    user_calibration_task = device.run(
        bell_circuit,
        gate_definitions=user_gate_calibrations.pulse_sequences,
        shots=num_shots,
        disable_qubit_rewiring=True,
    )
    device_calibration_task = device.run(
        bell_circuit,
        gate_definitions=device.gate_calibrations.pulse_sequences,
        shots=num_shots,
        disable_qubit_rewiring=True,
    )

    if not device.is_available:
        try:
            assert user_calibration_task.state not in AwsQuantumTask.TERMINAL_STATES
            assert device_calibration_task.state not in AwsQuantumTask.TERMINAL_STATES
        finally:
            user_calibration_task.cancel()
            device_calibration_task.cancel()
        return

    assert user_calibration_task.result().measurement_counts
    assert device_calibration_task.result().measurement_counts



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import functools

import numpy as np
import pytest

import braket.ir.jaqcd as ir
from braket.circuits import Circuit, FreeParameter, Gate, Instruction, QubitSet
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    QubitReferenceType,
)
from braket.ir.jaqcd.shared_models import (
    Angle,
    DoubleControl,
    DoubleTarget,
    MultiTarget,
    SingleControl,
    SingleTarget,
    TwoDimensionalMatrix,
)
from braket.pulse import ArbitraryWaveform, Frame, Port, PulseSequence


class NoTarget:
    pass


class TripleAngle:
    pass


class SingleNegControlModifier:
    pass


testdata = [
    (Gate.H, "h", ir.H, [SingleTarget], {}),
    (Gate.GPhase, "gphase", None, [NoTarget, Angle], {}),
    (Gate.I, "i", ir.I, [SingleTarget], {}),
    (Gate.X, "x", ir.X, [SingleTarget], {}),
    (Gate.Y, "y", ir.Y, [SingleTarget], {}),
    (Gate.Z, "z", ir.Z, [SingleTarget], {}),
    (Gate.S, "s", ir.S, [SingleTarget], {}),
    (Gate.Si, "si", ir.Si, [SingleTarget], {}),
    (Gate.T, "t", ir.T, [SingleTarget], {}),
    (Gate.Ti, "ti", ir.Ti, [SingleTarget], {}),
    (Gate.V, "v", ir.V, [SingleTarget], {}),
    (Gate.Vi, "vi", ir.Vi, [SingleTarget], {}),
    (Gate.Rx, "rx", ir.Rx, [SingleTarget, Angle], {}),
    (Gate.Ry, "ry", ir.Ry, [SingleTarget, Angle], {}),
    (Gate.Rz, "rz", ir.Rz, [SingleTarget, Angle], {}),
    (Gate.U, "u", None, [SingleTarget, TripleAngle], {}),
    (Gate.CNot, "cnot", ir.CNot, [SingleTarget, SingleControl], {}),
    (Gate.CV, "cv", ir.CV, [SingleTarget, SingleControl], {}),
    (Gate.CCNot, "ccnot", ir.CCNot, [SingleTarget, DoubleControl], {}),
    (Gate.Swap, "swap", ir.Swap, [DoubleTarget], {}),
    (Gate.CSwap, "cswap", ir.CSwap, [SingleControl, DoubleTarget], {}),
    (Gate.ISwap, "iswap", ir.ISwap, [DoubleTarget], {}),
    (Gate.PSwap, "pswap", ir.PSwap, [DoubleTarget, Angle], {}),
    (Gate.XY, "xy", ir.XY, [DoubleTarget, Angle], {}),
    (Gate.PhaseShift, "phaseshift", ir.PhaseShift, [SingleTarget, Angle], {}),
    (Gate.CPhaseShift, "cphaseshift", ir.CPhaseShift, [SingleControl, SingleTarget, Angle], {}),
    (
        Gate.CPhaseShift00,
        "cphaseshift00",
        ir.CPhaseShift00,
        [SingleControl, SingleTarget, Angle],
        {},
    ),
    (
        Gate.CPhaseShift01,
        "cphaseshift01",
        ir.CPhaseShift01,
        [SingleControl, SingleTarget, Angle],
        {},
    ),
    (
        Gate.CPhaseShift10,
        "cphaseshift10",
        ir.CPhaseShift10,
        [SingleControl, SingleTarget, Angle],
        {},
    ),
    (Gate.CY, "cy", ir.CY, [SingleTarget, SingleControl], {}),
    (Gate.CZ, "cz", ir.CZ, [SingleTarget, SingleControl], {}),
    (Gate.ECR, "ecr", ir.ECR, [DoubleTarget], {}),
    (Gate.XX, "xx", ir.XX, [DoubleTarget, Angle], {}),
    (Gate.YY, "yy", ir.YY, [DoubleTarget, Angle], {}),
    (Gate.ZZ, "zz", ir.ZZ, [DoubleTarget, Angle], {}),
    (Gate.GPi, "gpi", None, [SingleTarget, Angle], {}),
    (Gate.GPi2, "gpi2", None, [SingleTarget, Angle], {}),
    (Gate.MS, "ms", None, [DoubleTarget, TripleAngle], {}),
    (
        Gate.Unitary,
        "unitary",
        ir.Unitary,
        [TwoDimensionalMatrix, MultiTarget],
        {"input_type": complex},
    ),
    (
        Gate.Unitary,
        "unitary",
        ir.Unitary,
        [TwoDimensionalMatrix, MultiTarget],
        {"input_type": float},
    ),
    (
        Gate.Unitary,
        "unitary",
        ir.Unitary,
        [TwoDimensionalMatrix, MultiTarget],
        {"input_type": int},
    ),
]

parameterizable_gates = [
    Gate.GPhase,
    Gate.Rx,
    Gate.Ry,
    Gate.Rz,
    Gate.U,
    Gate.PhaseShift,
    Gate.PSwap,
    Gate.XX,
    Gate.XY,
    Gate.YY,
    Gate.ZZ,
    Gate.CPhaseShift,
    Gate.CPhaseShift00,
    Gate.CPhaseShift01,
    Gate.CPhaseShift10,
    Gate.GPi,
    Gate.GPi2,
    Gate.MS,
]

invalid_unitary_matrices = [
    (np.array([[1]])),
    (np.array([1])),
    (np.array([0, 1, 2])),
    (np.array([[0, 1], [1, 2], [3, 4]])),
    (np.array([[0, 1, 2], [2, 3]], dtype=object)),
    (np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])),
    (np.array([[0, 1], [1, 1]])),
]


def no_target_valid_input(**kwargs):
    return {}


def single_target_valid_input(**kwargs):
    return {"target": 2}


def double_target_valid_ir_input(**kwargs):
    return {"targets": [2, 3]}


def double_target_valid_input(**kwargs):
    return {"target1": 2, "target2": 3}


def angle_valid_input(**kwargs):
    return {"angle": 0.123}


def triple_angle_valid_input(**kwargs):
    return {"angle_1": 0.123, "angle_2": 4.567, "angle_3": 8.910}


def single_control_valid_input(**kwargs):
    return {"control": 0}


def single_neg_control_valid_input(**kwargs):
    return {"control": [0], "control_state": [0]}


def double_control_valid_ir_input(**kwargs):
    return {"controls": [0, 1]}


def double_control_valid_input(**kwargs):
    return {"control1": 0, "control2": 1}


def multi_target_valid_input(**kwargs):
    return {"targets": [5]}


def two_dimensional_matrix_valid_ir_input(**kwargs):
    return {"matrix": [[[0, 0], [1, 0]], [[1, 0], [0, 0]]]}


def two_dimensional_matrix_valid_input(**kwargs):
    input_type = kwargs.get("input_type")
    return {"matrix": np.array([[input_type(0), input_type(1)], [input_type(1), input_type(0)]])}


valid_ir_switcher = {
    "NoTarget": no_target_valid_input,
    "SingleTarget": single_target_valid_input,
    "DoubleTarget": double_target_valid_ir_input,
    "Angle": angle_valid_input,
    "TripleAngle": triple_angle_valid_input,
    "SingleControl": single_control_valid_input,
    "SingleNegControlModifier": single_neg_control_valid_input,
    "DoubleControl": double_control_valid_ir_input,
    "MultiTarget": multi_target_valid_input,
    "TwoDimensionalMatrix": two_dimensional_matrix_valid_ir_input,
}

valid_subroutine_switcher = dict(
    valid_ir_switcher,
    **{
        "TwoDimensionalMatrix": two_dimensional_matrix_valid_input,
        "DoubleTarget": double_target_valid_input,
        "DoubleControl": double_control_valid_input,
    },
)


def create_valid_ir_input(irsubclasses):
    input = {}
    for subclass in irsubclasses:
        input.update(valid_ir_switcher.get(subclass.__name__, lambda: "Invalid subclass")())
    return input


def create_valid_subroutine_input(irsubclasses, **kwargs):
    input = {}
    for subclass in irsubclasses:
        input.update(
            valid_subroutine_switcher.get(subclass.__name__, lambda: "Invalid subclass")(**kwargs)
        )
    return input


def create_valid_target_input(irsubclasses):
    input = {}
    qubit_set = []
    control_qubit_set = []
    control_state = None
    # based on the concept that control goes first in target input
    for subclass in irsubclasses:
        if subclass == NoTarget:
            qubit_set.extend(list(no_target_valid_input().values()))
        elif subclass == SingleTarget:
            qubit_set.extend(list(single_target_valid_input().values()))
        elif subclass == DoubleTarget:
            qubit_set.extend(list(double_target_valid_ir_input().values()))
        elif subclass == MultiTarget:
            qubit_set.extend(list(multi_target_valid_input().values()))
        elif subclass == SingleControl:
            qubit_set = list(single_control_valid_input().values()) + qubit_set
        elif subclass == SingleNegControlModifier:
            control_qubit_set = list(single_neg_control_valid_input()["control"])
            control_state = list(single_neg_control_valid_input()["control_state"])
        elif subclass == DoubleControl:
            qubit_set = list(double_control_valid_ir_input().values()) + qubit_set
        elif subclass in (Angle, TwoDimensionalMatrix, TripleAngle):
            pass
        else:
            raise ValueError("Invalid subclass")
    input["target"] = QubitSet(qubit_set)
    input["control"] = QubitSet(control_qubit_set)
    input["control_state"] = control_state
    return input


def create_valid_gate_class_input(irsubclasses, **kwargs):
    input = {}
    if Angle in irsubclasses:
        input.update(angle_valid_input())
    if TripleAngle in irsubclasses:
        input.update(triple_angle_valid_input())
    if TwoDimensionalMatrix in irsubclasses:
        input.update(two_dimensional_matrix_valid_input(**kwargs))
    return input


def create_valid_instruction_input(testclass, irsubclasses, **kwargs):
    input = create_valid_target_input(irsubclasses)
    input["operator"] = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs))
    return input


def calculate_qubit_count(irsubclasses):
    qubit_count = 0
    for subclass in irsubclasses:
        if subclass == SingleTarget:
            qubit_count += 1
        elif subclass == DoubleTarget:
            qubit_count += 2
        elif subclass == SingleControl:
            qubit_count += 1
        elif subclass == DoubleControl:
            qubit_count += 2
        elif subclass == MultiTarget:
            qubit_count += 3
        elif subclass in (NoTarget, Angle, TwoDimensionalMatrix, TripleAngle):
            pass
        else:
            raise ValueError("Invalid subclass")
    return qubit_count


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_ir_gate_level(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    if irclass is not None:
        expected = irclass(**create_valid_ir_input(irsubclasses))
        actual = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs)).to_ir(
            **create_valid_target_input(irsubclasses)
        )
        assert actual == expected


@pytest.mark.parametrize(
    "gate, target, serialization_properties, expected_ir",
    [
        (
            Gate.Rx(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "rx(0.17) q[4];",
        ),
        (
            Gate.Rx(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "rx(0.17) $4;",
        ),
        (
            Gate.X(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "x q[4];",
        ),
        (
            Gate.X(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "x $4;",
        ),
        (
            Gate.Z(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "z q[4];",
        ),
        (
            Gate.Z(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "z $4;",
        ),
        (
            Gate.Y(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "y q[4];",
        ),
        (
            Gate.Y(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "y $4;",
        ),
        (
            Gate.H(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "h q[4];",
        ),
        (
            Gate.H(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "h $4;",
        ),
        (
            Gate.Ry(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "ry(0.17) q[4];",
        ),
        (
            Gate.Ry(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "ry(0.17) $4;",
        ),
        (
            Gate.ZZ(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "zz(0.17) q[4], q[5];",
        ),
        (
            Gate.ZZ(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "zz(0.17) $4, $5;",
        ),
        (
            Gate.I(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "i q[4];",
        ),
        (
            Gate.I(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "i $4;",
        ),
        (
            Gate.V(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "v q[4];",
        ),
        (
            Gate.V(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "v $4;",
        ),
        (
            Gate.CY(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cy q[0], q[1];",
        ),
        (
            Gate.CY(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cy $0, $1;",
        ),
        (
            Gate.Rz(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "rz(0.17) q[4];",
        ),
        (
            Gate.Rz(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "rz(0.17) $4;",
        ),
        (
            Gate.U(angle_1=0.17, angle_2=3.45, angle_3=5.21),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "U(0.17, 3.45, 5.21) q[4];",
        ),
        (
            Gate.U(angle_1=0.17, angle_2=3.45, angle_3=5.21),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "U(0.17, 3.45, 5.21) $4;",
        ),
        (
            Gate.XX(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "xx(0.17) q[4], q[5];",
        ),
        (
            Gate.XX(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "xx(0.17) $4, $5;",
        ),
        (
            Gate.T(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "t q[4];",
        ),
        (
            Gate.T(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "t $4;",
        ),
        (
            Gate.CZ(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cz $0, $1;",
        ),
        (
            Gate.CZ(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cz q[0], q[1];",
        ),
        (
            Gate.YY(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "yy(0.17) q[4], q[5];",
        ),
        (
            Gate.YY(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "yy(0.17) $4, $5;",
        ),
        (
            Gate.XY(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "xy(0.17) q[4], q[5];",
        ),
        (
            Gate.XY(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "xy(0.17) $4, $5;",
        ),
        (
            Gate.ISwap(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "iswap $0, $1;",
        ),
        (
            Gate.ISwap(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "iswap q[0], q[1];",
        ),
        (
            Gate.Swap(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "swap $0, $1;",
        ),
        (
            Gate.Swap(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "swap q[0], q[1];",
        ),
        (
            Gate.ECR(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "ecr $0, $1;",
        ),
        (
            Gate.ECR(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "ecr q[0], q[1];",
        ),
        (
            Gate.CV(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cv $0, $1;",
        ),
        (
            Gate.CV(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cv q[0], q[1];",
        ),
        (
            Gate.Vi(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "vi q[4];",
        ),
        (
            Gate.Vi(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "vi $4;",
        ),
        (
            Gate.CSwap(),
            [0, 1, 2],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cswap q[0], q[1], q[2];",
        ),
        (
            Gate.CSwap(),
            [0, 1, 2],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cswap $0, $1, $2;",
        ),
        (
            Gate.CPhaseShift01(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cphaseshift01(0.17) q[4], q[5];",
        ),
        (
            Gate.CPhaseShift01(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cphaseshift01(0.17) $4, $5;",
        ),
        (
            Gate.CPhaseShift00(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cphaseshift00(0.17) q[4], q[5];",
        ),
        (
            Gate.CPhaseShift00(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cphaseshift00(0.17) $4, $5;",
        ),
        (
            Gate.CPhaseShift(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cphaseshift(0.17) q[4], q[5];",
        ),
        (
            Gate.CPhaseShift(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cphaseshift(0.17) $4, $5;",
        ),
        (
            Gate.S(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "s q[4];",
        ),
        (
            Gate.S(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "s $4;",
        ),
        (
            Gate.Si(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "si q[4];",
        ),
        (
            Gate.Si(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "si $4;",
        ),
        (
            Gate.Ti(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "ti q[4];",
        ),
        (
            Gate.Ti(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "ti $4;",
        ),
        (
            Gate.PhaseShift(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "phaseshift(0.17) q[4];",
        ),
        (
            Gate.PhaseShift(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "phaseshift(0.17) $4;",
        ),
        (
            Gate.CNot(),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cnot q[4], q[5];",
        ),
        (
            Gate.CNot(),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cnot $4, $5;",
        ),
        (
            Gate.PSwap(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "pswap(0.17) q[4], q[5];",
        ),
        (
            Gate.PSwap(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "pswap(0.17) $4, $5;",
        ),
        (
            Gate.CPhaseShift10(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cphaseshift10(0.17) q[4], q[5];",
        ),
        (
            Gate.CPhaseShift10(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cphaseshift10(0.17) $4, $5;",
        ),
        (
            Gate.CCNot(),
            [4, 5, 6],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "ccnot q[4], q[5], q[6];",
        ),
        (
            Gate.CCNot(),
            [4, 5, 6],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "ccnot $4, $5, $6;",
        ),
        (
            Gate.Unitary(Gate.CCNot().to_matrix()),
            [4, 5, 6],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket unitary(["
            "[1.0, 0, 0, 0, 0, 0, 0, 0], "
            "[0, 1.0, 0, 0, 0, 0, 0, 0], "
            "[0, 0, 1.0, 0, 0, 0, 0, 0], "
            "[0, 0, 0, 1.0, 0, 0, 0, 0], "
            "[0, 0, 0, 0, 1.0, 0, 0, 0], "
            "[0, 0, 0, 0, 0, 1.0, 0, 0], "
            "[0, 0, 0, 0, 0, 0, 0, 1.0], "
            "[0, 0, 0, 0, 0, 0, 1.0, 0]"
            "]) q[4], q[5], q[6]",
        ),
        (
            Gate.Unitary(Gate.CCNot().to_matrix()),
            [4, 5, 6],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "#pragma braket unitary(["
            "[1.0, 0, 0, 0, 0, 0, 0, 0], "
            "[0, 1.0, 0, 0, 0, 0, 0, 0], "
            "[0, 0, 1.0, 0, 0, 0, 0, 0], "
            "[0, 0, 0, 1.0, 0, 0, 0, 0], "
            "[0, 0, 0, 0, 1.0, 0, 0, 0], "
            "[0, 0, 0, 0, 0, 1.0, 0, 0], "
            "[0, 0, 0, 0, 0, 0, 0, 1.0], "
            "[0, 0, 0, 0, 0, 0, 1.0, 0]"
            "]) $4, $5, $6",
        ),
        (
            Gate.Unitary(np.round(Gate.ECR().to_matrix(), 8)),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket unitary(["
            "[0, 0, 0.70710678, 0.70710678im], "
            "[0, 0, 0.70710678im, 0.70710678], "
            "[0.70710678, -0.70710678im, 0, 0], "
            "[-0.70710678im, 0.70710678, 0, 0]"
            "]) q[4], q[5]",
        ),
        (
            Gate.Unitary(np.round(Gate.ECR().to_matrix(), 8)),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "#pragma braket unitary(["
            "[0, 0, 0.70710678, 0.70710678im], "
            "[0, 0, 0.70710678im, 0.70710678], "
            "[0.70710678, -0.70710678im, 0, 0], "
            "[-0.70710678im, 0.70710678, 0, 0]"
            "]) $4, $5",
        ),
        (
            Gate.Unitary(np.round(Gate.T().to_matrix(), 8)),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket unitary([[1.0, 0], [0, 0.70710678 + 0.70710678im]]) q[4]",
        ),
        (
            Gate.Unitary(np.round(Gate.T().to_matrix(), 8)),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "#pragma braket unitary([[1.0, 0], [0, 0.70710678 + 0.70710678im]]) $4",
        ),
        (
            Gate.Unitary(np.array([[1.0, 0], [0, 0.70710678 - 0.70710678j]])),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket unitary([[1.0, 0], [0, 0.70710678 - 0.70710678im]]) q[4]",
        ),
        (
            Gate.PulseGate(
                PulseSequence().play(
                    Frame("user_frame", Port("device_port_x", 1e-9), 1e9),
                    ArbitraryWaveform([1, 2], "arb_wf"),
                ),
                1,
            ),
            [0],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "\n".join(["cal {", "    play(user_frame, arb_wf);", "}"]),
        ),
        (
            Gate.GPi(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "gpi(0.17) q[4];",
        ),
        (
            Gate.GPi(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "gpi(0.17) $4;",
        ),
        (
            Gate.GPi2(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "gpi2(0.17) q[4];",
        ),
        (
            Gate.GPi2(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "gpi2(0.17) $4;",
        ),
        (
            Gate.MS(angle_1=0.17, angle_2=3.45),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            f"ms(0.17, 3.45, {np.pi / 2}) q[4], q[5];",
        ),
        (
            Gate.MS(angle_1=0.17, angle_2=3.45),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            f"ms(0.17, 3.45, {np.pi / 2}) $4, $5;",
        ),
    ],
)
def test_gate_to_ir_openqasm(gate, target, serialization_properties, expected_ir):
    assert (
        gate.to_ir(
            target, ir_type=IRType.OPENQASM, serialization_properties=serialization_properties
        )
        == expected_ir
    )


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_ir_instruction_level(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    if irclass is not None:
        expected = irclass(**create_valid_ir_input(irsubclasses))
        instruction = Instruction(
            **create_valid_instruction_input(testclass, irsubclasses, **kwargs)
        )
        actual = instruction.to_ir()
        assert actual == expected


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_gate_subroutine(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    qubit_count = calculate_qubit_count(irsubclasses)
    subroutine = getattr(Circuit(), subroutine_name)
    assert subroutine(**create_valid_subroutine_input(irsubclasses, **kwargs)) == Circuit(
        Instruction(**create_valid_instruction_input(testclass, irsubclasses, **kwargs))
    )
    if qubit_count == 1:
        multi_targets = [0, 1, 2]
        instruction_list = []
        for target in multi_targets:
            instruction_list.append(
                Instruction(
                    operator=testclass(**create_valid_gate_class_input(irsubclasses, **kwargs)),
                    target=target,
                )
            )
        subroutine = getattr(Circuit(), subroutine_name)
        subroutine_input = {"target": multi_targets}
        if Angle in irsubclasses:
            subroutine_input.update(angle_valid_input())
        if TripleAngle in irsubclasses:
            subroutine_input.update(triple_angle_valid_input())
        assert subroutine(**subroutine_input) == Circuit(instruction_list)


@pytest.mark.parametrize(
    "control, control_state, instruction_set",
    [
        (
            2,
            None,
            Instruction(**create_valid_instruction_input(Gate.PhaseShift, [SingleTarget, Angle])),
        ),
        (
            2,
            [0],
            [
                Instruction(operator=Gate.X(), target=2),
                Instruction(
                    **create_valid_instruction_input(Gate.PhaseShift, [SingleTarget, Angle])
                ),
                Instruction(operator=Gate.X(), target=2),
            ],
        ),
        (
            [0, 2],
            [0, 1],
            Instruction(
                **create_valid_instruction_input(
                    Gate.PhaseShift, [SingleTarget, SingleNegControlModifier, Angle]
                )
            ),
        ),
    ],
)
def test_control_gphase_subroutine(control, control_state, instruction_set):
    subroutine = getattr(Circuit(), "gphase")
    assert subroutine(angle=0.123, control=control, control_state=control_state) == Circuit(
        instruction_set
    )


def test_angle_gphase_is_none():
    with pytest.raises(ValueError, match="angle must not be None"):
        Gate.GPhase(angle=None)


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_gate_adjoint_expansion_correct(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    gate = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs))
    matrices = [elem.to_matrix() for elem in gate.adjoint()]
    matrices.append(gate.to_matrix())
    identity = np.eye(2**gate.qubit_count)
    assert np.allclose(functools.reduce(lambda a, b: a @ b, matrices), identity)


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_gate_to_matrix(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    gate1 = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs))
    gate2 = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs))
    assert isinstance(gate1.to_matrix(), np.ndarray)
    assert gate1.matrix_equivalence(gate2)


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_fixed_qubit_count(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    fixed_qubit_count = testclass.fixed_qubit_count()
    if fixed_qubit_count is not NotImplemented:
        gate = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs))
        assert gate.qubit_count == fixed_qubit_count


# Additional Unitary gate tests


def test_equality():
    u1 = Gate.Unitary(np.array([[0 + 0j, 1 + 0j], [1 + 0j, 0 + 0j]]))
    u2 = Gate.Unitary(np.array([[0, 1], [1, 0]], dtype=np.float32), display_name=["u2"])
    other_gate = Gate.Unitary(np.array([[1, 0], [0, 1]]))
    non_gate = "non gate"

    assert u1 == u2
    assert u1 is not u2
    assert u1 != other_gate
    assert u1 != non_gate


def test_free_param_equality():
    param1 = FreeParameter("theta")
    param2 = FreeParameter("phi")
    rx1 = Gate.Rx(param1)
    rx2 = Gate.Rx(param1)
    other_gate = Gate.Rx(param2)

    assert rx1 == rx2
    assert rx1 is not rx2
    assert rx1 != other_gate
    assert rx1 != param1


def test_large_unitary():
    matrix = np.eye(16, dtype=np.float32)
    # Permute rows of matrix
    matrix[[*range(16)]] = matrix[[(i + 1) % 16 for i in range(16)]]
    unitary = Gate.Unitary(matrix)
    assert unitary.qubit_count == 4


@pytest.mark.parametrize("gate", parameterizable_gates)
def test_bind_values(gate):
    triple_angled = gate.__name__ in ("MS", "U")
    num_params = 3 if triple_angled else 1
    thetas = [FreeParameter(f"theta_{i}") for i in range(num_params)]
    mapping = {f"theta_{i}": i for i in range(num_params)}
    param_gate = gate(*thetas)
    new_gate = param_gate.bind_values(**mapping)
    expected = gate(*range(num_params))

    assert type(new_gate) is type(param_gate) and new_gate == expected
    if triple_angled:
        for angle in new_gate.angle_1, new_gate.angle_2, new_gate.angle_3:
            assert isinstance(angle, float)
    else:
        assert isinstance(new_gate.angle, float)


def test_bind_values_pulse_gate():
    qubit_count = 1
    frame = Frame("user_frame", Port("device_port_x", 1e-9), 1e9)
    gate = Gate.PulseGate(
        PulseSequence()
        .set_frequency(frame, FreeParameter("a") + FreeParameter("b"))
        .delay(frame, FreeParameter("c")),
        qubit_count,
    )

    def to_ir(pulse_gate):
        return pulse_gate.to_ir(range(pulse_gate.qubit_count), IRType.OPENQASM)

    a = 3
    a_bound = gate.bind_values(a=a)
    a_bound_ir = to_ir(a_bound)

    assert a_bound_ir == "\n".join(
        [
            "cal {",
            "    set_frequency(user_frame, 3.0 + b);",
            "    delay[c * 1s] user_frame;",
            "}",
        ]
    )

    assert a_bound_ir == to_ir(
        Gate.PulseGate(gate.pulse_sequence.make_bound_pulse_sequence({"a": a}), qubit_count)
    )
    assert a_bound_ir != to_ir(gate)

    c = 4e-6
    ac_bound = a_bound.bind_values(c=c)
    ac_bound_ir = to_ir(ac_bound)
    assert ac_bound_ir == to_ir(
        Gate.PulseGate(a_bound.pulse_sequence.make_bound_pulse_sequence({"c": c}), qubit_count)
    )
    assert ac_bound_ir != a_bound_ir


def test_pulse_gate_capture_throws():
    with pytest.raises(ValueError):
        Circuit().pulse_gate(
            0, PulseSequence().capture_v0(Frame("user_frame", Port("device_port_x", dt=1e-9), 1e9))
        )


@pytest.mark.parametrize("matrix", invalid_unitary_matrices)
def test_unitary_invalid_matrix(matrix):
    with pytest.raises(ValueError):
        Gate.Unitary(matrix=matrix)


def test_unitary_matrix_target_size_mismatch():
    with pytest.raises(ValueError):
        Circuit().unitary(
            matrix=np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]), targets=[0]
        )


def test_pulse_gate_to_matrix():
    with pytest.raises(NotImplementedError):
        Gate.PulseGate(
            PulseSequence().play(
                Frame("user_frame", Port("device_port_x", 1e-9), 1e9),
                ArbitraryWaveform([1, 2], "arb_wf"),
            ),
            1,
        ).to_matrix()


@pytest.mark.parametrize(
    "gate, target, control, control_state, expected_ir",
    (
        (Gate.H(), QubitSet(0), QubitSet(1), None, "ctrl @ h q[1], q[0];"),
        (Gate.H(), QubitSet(0), QubitSet([1, 2]), None, "ctrl(2) @ h q[1], q[2], q[0];"),
        (Gate.Ry(angle=1.23), QubitSet(0), QubitSet([2]), None, "ctrl @ ry(1.23) q[2], q[0];"),
        (
            Gate.MS(angle_1=0.17, angle_2=3.45),
            QubitSet(0),
            QubitSet([1, 2]),
            None,
            f"ctrl(2) @ ms(0.17, 3.45, {np.pi / 2}) q[1], q[2], q[0];",
        ),
        (
            Gate.CCNot(),
            QubitSet([0, 1, 2]),
            QubitSet([3, 4]),
            None,
            "ctrl(2) @ ccnot q[3], q[4], q[0], q[1], q[2];",
        ),
        (
            Gate.Z(),
            QubitSet([0]),
            QubitSet([1, 2, 3, 4, 5, 6, 7]),
            [1, 1, 1, 0, 0, 1, 0],
            "ctrl(3) @ negctrl(2) @ ctrl @ negctrl @ "
            "z q[1], q[2], q[3], q[4], q[5], q[6], q[7], q[0];",
        ),
        (
            Gate.Z(),
            QubitSet([0]),
            QubitSet([1, 2, 3, 4, 5, 6, 7]),
            "1110010",
            "ctrl(3) @ negctrl(2) @ ctrl @ negctrl @ "
            "z q[1], q[2], q[3], q[4], q[5], q[6], q[7], q[0];",
        ),
        (
            Gate.Z(),
            QubitSet([0]),
            QubitSet([1, 2, 3, 4, 5, 6, 7]),
            114,
            "ctrl(3) @ negctrl(2) @ ctrl @ negctrl @ "
            "z q[1], q[2], q[3], q[4], q[5], q[6], q[7], q[0];",
        ),
        (
            Gate.Z(),
            QubitSet([0]),
            QubitSet([1, 2, 3]),
            [1, 0],
            "negctrl @ ctrl @ negctrl @ z q[1], q[2], q[3], q[0];",
        ),
        (
            Gate.Z(),
            QubitSet([0]),
            QubitSet([1, 2, 3]),
            "10",
            "negctrl @ ctrl @ negctrl @ z q[1], q[2], q[3], q[0];",
        ),
        (
            Gate.GPhase(0.3),
            QubitSet([]),
            QubitSet([1]),
            "1",
            "ctrl @ gphase(0.3) q[1];",
        ),
    ),
)
def test_gate_control(gate, target, control, control_state, expected_ir):
    serialization_properties = OpenQASMSerializationProperties(
        qubit_reference_type=QubitReferenceType.VIRTUAL
    )
    assert (
        gate.to_ir(
            target,
            control=control,
            control_state=control_state,
            ir_type=IRType.OPENQASM,
            serialization_properties=serialization_properties,
        )
        == expected_ir
    )


@pytest.mark.parametrize(
    "control, control_state, error_string",
    (
        (
            [0, 1],
            [1, 0, 1],
            "State value represents a binary sequence of length greater "
            "than the specified number of qubits.",
        ),
        (
            [0, 1],
            "101",
            "State value represents a binary sequence of length greater "
            "than the specified number of qubits.",
        ),
        (
            [0, 1],
            5,
            "State value represents a binary sequence of length greater "
            "than the specified number of qubits.",
        ),
    ),
)
def test_gate_control_invalid_state(control, control_state, error_string):
    with pytest.raises(ValueError, match=error_string):
        Gate.X().to_ir(
            target=[0],
            control=control,
            control_state=control_state,
            ir_type=IRType.OPENQASM,
            serialization_properties=OpenQASMSerializationProperties(
                qubit_reference_type=QubitReferenceType.VIRTUAL
            ),
        )


@pytest.mark.parametrize(
    "gate, target, power, expected_ir",
    (
        (Gate.H(), QubitSet(0), 2, "pow(2) @ h q[0];"),
        (Gate.H(), QubitSet(0), 2.0, "pow(2.0) @ h q[0];"),
        (Gate.H(), QubitSet(0), 2.5, "pow(2.5) @ h q[0];"),
        (Gate.H(), QubitSet(0), 0, "pow(0) @ h q[0];"),
        (Gate.H(), QubitSet(0), -2, "inv @ pow(2) @ h q[0];"),
        (Gate.H(), QubitSet(0), -2.0, "inv @ pow(2.0) @ h q[0];"),
        (Gate.H(), QubitSet(0), -2.5, "inv @ pow(2.5) @ h q[0];"),
    ),
)
def test_gate_power(gate, target, power, expected_ir):
    serialization_properties = OpenQASMSerializationProperties(
        qubit_reference_type=QubitReferenceType.VIRTUAL
    )
    assert (
        gate.to_ir(
            target,
            power=power,
            ir_type=IRType.OPENQASM,
            serialization_properties=serialization_properties,
        )
        == expected_ir
    )


def test_hash():
    assert hash(Gate.Unitary(Gate.CCNot().to_matrix())) == hash(
        Gate.Unitary(Gate.CCNot().to_matrix())
    )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from unittest.mock import Mock

from braket.tracking.tracking_context import (
    active_trackers,
    broadcast_event,
    deregister_tracker,
    register_tracker,
)


def test_tracking_context():
    assert active_trackers() == set()


def test_register_deregister_tracker():
    register_tracker("foo")
    assert active_trackers() == {"foo"}
    register_tracker("bar")
    register_tracker("bar")
    assert active_trackers() == {"foo", "bar"}
    deregister_tracker("foo")
    assert active_trackers() == {"bar"}
    deregister_tracker("bar")


def test_broadcast_event():
    tracker = Mock()
    register_tracker(tracker)
    broadcast_event("EVENT")
    tracker.receive_event.assert_called_with("EVENT")
    deregister_tracker(tracker)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import collections
import os
import sys
from collections.abc import Generator

##############################################################################
#
# Support for reading logs
#
##############################################################################
from typing import ClassVar, Optional

from botocore.exceptions import ClientError

from braket.aws.aws_session import AwsSession


class ColorWrap:
    """A callable that prints text in a different color depending on the instance.
    Up to 5 if the standard output is a terminal or a Jupyter notebook cell.
    """

    # For what color each number represents, see
    # https://misc.flogisoft.com/bash/tip_colors_and_formatting#colors
    _stream_colors: ClassVar = [34, 35, 32, 36, 33]

    def __init__(self, force: bool = False):
        """Initialize a `ColorWrap`.

        Args:
            force (bool): If True, the render output is colorized wherever the
                output is. Default: False.
        """
        self.colorize = force or sys.stdout.isatty() or os.environ.get("JPY_PARENT_PID", None)

    def __call__(self, index: int, s: str):
        """Prints the string, colorized or not, depending on the environment.

        Args:
            index (int): The instance number.
            s (str): The string to print.
        """
        if self.colorize:
            self._color_wrap(index, s)
        else:
            print(s)

    def _color_wrap(self, index: int, s: str) -> None:
        """Prints the string in a color determined by the index.

        Args:
            index (int): The instance number.
            s (str): The string to print (color-wrapped).
        """
        print(f"\x1b[{self._stream_colors[index % len(self._stream_colors)]}m{s}\x1b[0m")


# Position is a tuple that includes the last read timestamp and the number of items that were read
# at that time. This is used to figure out which event to start with on the next read.
Position = collections.namedtuple("Position", ["timestamp", "skip"])


def multi_stream_iter(
    aws_session: AwsSession, log_group: str, streams: list[str], positions: dict[str, Position]
) -> Generator[tuple[int, dict]]:
    """Iterates over the available events coming from a set of log streams.
    Log streams are in a single log group interleaving the events from each stream,
    so they yield in timestamp order.

    Args:
        aws_session (AwsSession): The AwsSession for interfacing with CloudWatch.
        log_group (str): The name of the log group.
        streams (list[str]): A list of the log stream names. The the stream number is
            the position of the stream in this list.
        positions (dict[str, Position]): A list of (timestamp, skip) pairs which represent
            the last record read from each stream.

    Yields:
        Generator[tuple[int, dict]]: A tuple of (stream number, cloudwatch log event).
    """
    event_iters = [
        log_stream(aws_session, log_group, s, positions[s].timestamp, positions[s].skip)
        for s in streams
    ]
    events = []
    for s in event_iters:
        try:
            events.append(next(s))
        except StopIteration:
            events.append(None)

    while any(events):
        i = events.index(min(events, key=lambda x: x["timestamp"] if x else float("inf")))
        yield i, events[i]
        try:
            events[i] = next(event_iters[i])
        except StopIteration:
            events[i] = None


def log_stream(
    aws_session: AwsSession, log_group: str, stream_name: str, start_time: int = 0, skip: int = 0
) -> Generator[dict]:
    """A generator for log items in a single stream.
    This yields all the items that are available at the current moment.

    Args:
        aws_session (AwsSession): The AwsSession for interfacing with CloudWatch.
        log_group (str): The name of the log group.
        stream_name (str): The name of the specific stream.
        start_time (int): The time stamp value to start reading the logs from. Default: 0.
        skip (int): The number of log entries to skip at the start. Default: 0 (This is for
            when there are multiple entries at the same timestamp.)

    Yields:
        Generator[dict]: A CloudWatch log event with the following key-value pairs:
        'timestamp' (int): The time of the event.
        'message' (str): The log event data.
        'ingestionTime' (int): The time the event was ingested.
    """
    next_token = None

    event_count = 1
    while event_count > 0:
        response = aws_session.get_log_events(
            log_group,
            stream_name,
            start_time,
            start_from_head=True,
            next_token=next_token,
        )
        next_token = response["nextForwardToken"]
        events = response["events"]
        event_count = len(events)
        if event_count > skip:
            events = events[skip:]
            skip = 0
        else:
            skip = skip - event_count
            events = []
        yield from events


def flush_log_streams(  # noqa C901
    aws_session: AwsSession,
    log_group: str,
    stream_prefix: str,
    stream_names: list[str],
    positions: dict[str, Position],
    stream_count: int,
    has_streams: bool,
    color_wrap: ColorWrap,
    state: list[str],
    queue_position: Optional[str] = None,
) -> bool:
    """Flushes log streams to stdout.

    Args:
        aws_session (AwsSession): The AwsSession for interfacing with CloudWatch.
        log_group (str): The name of the log group.
        stream_prefix (str): The prefix for log streams to flush.
        stream_names (list[str]): A list of the log stream names. The position of the stream in
            this list is the stream number. If incomplete, the function will check for remaining
            streams and mutate this list to add stream names when available, up to the
            `stream_count` limit.
        positions (dict[str, Position]): A dict mapping stream numbers to (timestamp, skip) pairs
            which represent the last record read from each stream. The function will update this
            list after being called to represent the new last record read from each stream.
        stream_count (int): The number of streams expected.
        has_streams (bool): Whether the function has already been called once all streams have
            been found. This value is possibly updated and returned at the end of execution.
        color_wrap (ColorWrap): An instance of ColorWrap to potentially color-wrap print statements
            from different streams.
        state (list[str]): The previous and current state of the job.
        queue_position (Optional[str]): The current queue position. This is not passed in if the job
            is ran with `quiet=True`

    Raises:
        Exception: Any exception found besides a ResourceNotFoundException.

    Returns:
        bool: Returns 'True' if any streams have been flushed.
    """
    if len(stream_names) < stream_count:
        # Log streams are created whenever a container starts writing to stdout/err,
        # so this list may be dynamic until we have a stream for every instance.
        try:
            streams = aws_session.describe_log_streams(
                log_group,
                stream_prefix,
                limit=stream_count,
            )
            # stream_names = [...] wouldn't modify the list by reference.
            new_streams = [
                s["logStreamName"]
                for s in streams["logStreams"]
                if s["logStreamName"] not in stream_names
            ]
            stream_names.extend(new_streams)
            positions.update(
                [(s, Position(timestamp=0, skip=0)) for s in stream_names if s not in positions]
            )
        except ClientError as e:
            # On the very first training job run on an account, there's no
            # log group until the container starts logging, so ignore any
            # errors thrown about that until logging begins.
            err = e.response.get("Error", {})
            if err.get("Code") != "ResourceNotFoundException":
                raise

    if len(stream_names) > 0:
        if not has_streams:
            print()
            has_streams = True
        for idx, event in multi_stream_iter(aws_session, log_group, stream_names, positions):
            color_wrap(idx, event["message"])
            ts, count = positions[stream_names[idx]]
            if event["timestamp"] == ts:
                positions[stream_names[idx]] = Position(timestamp=ts, skip=count + 1)
            else:
                positions[stream_names[idx]] = Position(timestamp=event["timestamp"], skip=1)
    elif queue_position is not None and state[1] == "QUEUED":
        print(f"Job queue position: {queue_position}", end="\n", flush=True)
    elif state[0] != state[1] and state[1] == "RUNNING" and queue_position is not None:
        print("Running:", end="\n", flush=True)
    else:
        print(".", end="", flush=True)
    return has_streams



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from typing import Union

from braket.ahs.discretization_types import DiscretizationProperties
from braket.ahs.field import Field
from braket.ahs.hamiltonian import Hamiltonian
from braket.timings.time_series import StitchBoundaryCondition, TimeSeries


class DrivingField(Hamiltonian):
    def __init__(
        self,
        amplitude: Union[Field, TimeSeries],
        phase: Union[Field, TimeSeries],
        detuning: Union[Field, TimeSeries],
    ) -> None:
        r"""Creates a Hamiltonian term :math:`H_{drive}` for the driving field
        that coherently transfers atoms from the ground state to the Rydberg state
        in an AnalogHamiltonianSimulation, defined by the formula

        .. math::
            H_{drive} (t) := \frac{\Omega(t)}{2} e^{i \phi(t)} \left(
                \sum_k |g_k \rangle \langle r_k| + |r_k \rangle \langle g_k|
            \right) - \Delta(t) \sum_k{| r_k \rangle \langle r_k |}

        where

            :math:`\Omega(t)` is the global Rabi frequency in rad/s,

            :math:`\phi(t)` is the global phase in rad/s,

            :math:`\Delta(t)` is the global detuning in rad/s,

            :math:`|g_k \rangle` is the ground state of atom :math:`k`,

            :math:`|r_k \rangle` is the Rydberg state of atom :math:`k`.

        with the sum :math:`\sum_k` taken over all target atoms.

        Args:
            amplitude (Union[Field, TimeSeries]): global amplitude (:math:`\Omega(t)`).
                Time is in s, and value is in rad/s.
            phase (Union[Field, TimeSeries]): global phase (:math:`\phi(t)`).
                Time is in s, and value is in rad/s.
            detuning (Union[Field, TimeSeries]): global detuning (:math:`\Delta(t)`).
                Time is in s, and value is in rad/s.
        """
        super().__init__()
        self._amplitude = amplitude if isinstance(amplitude, Field) else Field(amplitude)
        self._phase = phase if isinstance(phase, Field) else Field(phase)
        self._detuning = detuning if isinstance(detuning, Field) else Field(detuning)

    @property
    def terms(self) -> list[Hamiltonian]:
        return [self]

    @property
    def amplitude(self) -> Field:
        r"""Field: The global amplitude (:math:`\Omega(t)`). Time is in s, and value is in rad/s."""
        return self._amplitude

    @property
    def phase(self) -> Field:
        r"""Field: The global phase (:math:`\phi(t)`). Time is in s, and value is in rad/s."""
        return self._phase

    @property
    def detuning(self) -> Field:
        r"""Field: global detuning (:math:`\Delta(t)`). Time is in s, and value is in rad/s."""
        return self._detuning

    def stitch(
        self, other: DrivingField, boundary: StitchBoundaryCondition = StitchBoundaryCondition.MEAN
    ) -> DrivingField:
        """Stitches two driving fields based on TimeSeries.stitch method.
        The time points of the second DrivingField are shifted such that the first time point of
        the second DrifingField coincides with the last time point of the first DrivingField.
        The boundary point value is handled according to StitchBoundaryCondition argument value.

        Args:
            other (DrivingField): The second shifting field to be stitched with.
            boundary (StitchBoundaryCondition): {"mean", "left", "right"}. Boundary point handler.

                Possible options are
              - "mean" - take the average of the boundary value points of the first
                and the second time series.
              - "left" - use the last value from the left time series as the boundary point.
              - "right" - use the first value from the right time series as the boundary point.

        Returns:
            DrivingField: The stitched DrivingField object.
        """
        amplitude = self.amplitude.time_series.stitch(other.amplitude.time_series, boundary)
        detuning = self.detuning.time_series.stitch(other.detuning.time_series, boundary)
        phase = self.phase.time_series.stitch(other.phase.time_series, boundary)

        return DrivingField(amplitude=amplitude, detuning=detuning, phase=phase)

    def discretize(self, properties: DiscretizationProperties) -> DrivingField:
        """Creates a discretized version of the Hamiltonian.

        Args:
            properties (DiscretizationProperties): Capabilities of a device that represent the
                resolution with which the device can implement the parameters.

        Returns:
            DrivingField: A new discretized DrivingField.
        """
        driving_parameters = properties.rydberg.rydbergGlobal
        time_resolution = driving_parameters.timeResolution
        discretized_amplitude = self.amplitude.discretize(
            time_resolution=time_resolution,
            value_resolution=driving_parameters.rabiFrequencyResolution,
        )
        discretized_phase = self.phase.discretize(
            time_resolution=time_resolution,
            value_resolution=driving_parameters.phaseResolution,
        )
        discretized_detuning = self.detuning.discretize(
            time_resolution=time_resolution,
            value_resolution=driving_parameters.detuningResolution,
        )
        return DrivingField(
            amplitude=discretized_amplitude, phase=discretized_phase, detuning=discretized_detuning
        )

    @staticmethod
    def from_lists(
        times: list[float], amplitudes: list[float], detunings: list[float], phases: list[float]
    ) -> DrivingField:
        """Builds DrivingField Hamiltonian from lists defining time evolution
        of Hamiltonian parameters (Rabi frequency, detuning, phase).
        The values of the parameters at each time points are global for all atoms.

        Args:
            times (list[float]): The time points of the driving field
            amplitudes (list[float]): The values of the amplitude
            detunings (list[float]): The values of the detuning
            phases (list[float]): The values of the phase

        Raises:
            ValueError: If any of the input args length is different from the rest.

        Returns:
            DrivingField: DrivingField Hamiltonian.
        """
        if not (len(times) == len(amplitudes) == len(detunings) == len(phases)):
            raise ValueError(
                f"The lengths of the lists for times({len(times)}), amplitudes({len(amplitudes)}),\
                detunings({len(detunings)}) and phases({len(phases)}) are not equal"
            )

        amplitude = TimeSeries()
        detuning = TimeSeries()
        phase = TimeSeries()

        for t, amplitude_value, detuning_value, phase_value in zip(
            times, amplitudes, detunings, phases
        ):
            amplitude.put(t, amplitude_value)
            detuning.put(t, detuning_value)
            phase.put(t, phase_value)

        drive = DrivingField(amplitude=amplitude, detuning=detuning, phase=phase)

        return drive



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import itertools
from typing import Optional, Union

from braket.circuits.circuit import Circuit
from braket.circuits.observables import I, TensorProduct, X, Y, Z

_IDENTITY = "I"
_PAULI_X = "X"
_PAULI_Y = "Y"
_PAULI_Z = "Z"
_PAULI_INDICES = {_IDENTITY: 0, _PAULI_X: 1, _PAULI_Y: 2, _PAULI_Z: 3}
_PRODUCT_MAP = {
    "X": {"Y": ["Z", 1j], "Z": ["Y", -1j]},
    "Y": {"X": ["Z", -1j], "Z": ["X", 1j]},
    "Z": {"X": ["Y", 1j], "Y": ["X", -1j]},
}
_ID_OBS = I()
_PAULI_OBSERVABLES = {_PAULI_X: X(), _PAULI_Y: Y(), _PAULI_Z: Z()}
_SIGN_MAP = {"+": 1, "-": -1}


class PauliString:
    """A lightweight representation of a Pauli string with its phase."""

    def __init__(self, pauli_string: Union[str, PauliString]):
        """Initializes a `PauliString`.

        Args:
            pauli_string (Union[str, PauliString]): The representation of the pauli word, either a
                string or another PauliString object. A valid string consists of an optional phase,
                specified by an optional sign +/- followed by an uppercase string in {I, X, Y, Z}.
                Example valid strings are: XYZ, +YIZY, -YX

        Raises:
            ValueError: If the Pauli String is empty.
        """
        if not pauli_string:
            raise ValueError("pauli_string must not be empty")
        if isinstance(pauli_string, PauliString):
            self._phase = pauli_string._phase
            self._qubit_count = pauli_string._qubit_count
            self._nontrivial = pauli_string._nontrivial
        elif isinstance(pauli_string, str):
            self._phase, factors_str = PauliString._split(pauli_string)
            self._qubit_count = len(factors_str)
            self._nontrivial = {
                i: factors_str[i] for i in range(len(factors_str)) if factors_str[i] != "I"
            }
        else:
            raise TypeError(f"Pauli word {pauli_string} must be of type {PauliString} or {str}")

    @property
    def phase(self) -> int:
        """int: The phase of the Pauli string.

        Can be one of +/-1
        """
        return self._phase

    @property
    def qubit_count(self) -> int:
        """int: The number of qubits this Pauli string acts on."""
        return self._qubit_count

    def to_unsigned_observable(self, include_trivial: bool = False) -> TensorProduct:
        """Returns the observable corresponding to the unsigned part of the Pauli string.

        For example, for a Pauli string -XYZ, the corresponding observable is X ⊗ Y ⊗ Z.

        Args:
            include_trivial (bool): Whether to include explicit identity factors in the observable.
                Default: False.

        Returns:
            TensorProduct: The tensor product of the unsigned factors in the Pauli string.
        """
        if include_trivial:
            return TensorProduct(
                [
                    (
                        _PAULI_OBSERVABLES[self._nontrivial[qubit]]
                        if qubit in self._nontrivial
                        else _ID_OBS
                    )
                    for qubit in range(self._qubit_count)
                ]
            )
        return TensorProduct(
            [_PAULI_OBSERVABLES[self._nontrivial[qubit]] for qubit in sorted(self._nontrivial)]
        )

    def weight_n_substrings(self, weight: int) -> tuple[PauliString, ...]:
        r"""Returns every substring of this Pauli string with exactly `weight` nontrivial factors.

        The number of substrings is equal to :math:`\binom{n}{w}`, where :math`n` is the number of
        nontrivial (non-identity) factors in the Pauli string and :math`w` is `weight`.

        Args:
            weight (int): The number of non-identity factors in the substrings.

        Returns:
            tuple[PauliString, ...]: A tuple of weight-n Pauli substrings.
        """
        substrings = []
        for indices in itertools.combinations(self._nontrivial, weight):
            factors = [
                (
                    self._nontrivial[qubit]
                    if qubit in set(indices).intersection(self._nontrivial)
                    else "I"
                )
                for qubit in range(self._qubit_count)
            ]
            substrings.append(
                PauliString(f"{PauliString._phase_to_str(self._phase)}{''.join(factors)}")
            )
        return tuple(substrings)

    def eigenstate(self, signs: Optional[Union[str, list[int], tuple[int, ...]]] = None) -> Circuit:
        """Returns the eigenstate of this Pauli string with the given factor signs.

        The resulting eigenstate has each qubit in the +1 eigenstate of its corresponding signed
        Pauli operator. For example, a Pauli string +XYZ and signs ++- has factors +X, +Y and -Z,
        with the corresponding qubits in states `|+⟩` , `|i⟩` , and `|1⟩` respectively (the global
        phase of the Pauli string is ignored).

        Args:
            signs (Optional[Union[str, list[int], tuple[int, ...]]]): The sign of each factor of the
                eigenstate, specified either as a string of "+" and "_", or as a list or tuple of
                +/-1. The length of signs must be equal to the length of the Pauli string. If not
                specified, it is assumed to be all +. Default: None.

        Returns:
            Circuit: A circuit that prepares the desired eigenstate of the Pauli string.

        Raises:
            ValueError: If the length of signs is not equal to that of the Pauli string or the signs
                are invalid.
        """
        qubit_count = self._qubit_count
        if not signs:
            signs = "+" * qubit_count
        elif len(signs) != qubit_count:
            raise ValueError(
                f"signs must be the same length of the Pauli string ({qubit_count}), "
                f"but was {len(signs)}"
            )
        signs_tup = (
            tuple(_SIGN_MAP.get(sign) for sign in signs) if isinstance(signs, str) else tuple(signs)
        )
        if not set(signs_tup) <= {1, -1}:
            raise ValueError(f"signs must be +/-1, got {signs}")
        return self._generate_eigenstate_circuit(signs_tup)

    def dot(self, other: PauliString, inplace: bool = False) -> PauliString:
        """Right multiplies this Pauli string with the argument.

        Returns the result of multiplying the current circuit by the argument on its right. For
        example, if called on `-XYZ` with argument `ZYX`, then `YIY` is the result. In-place
        computation is off by default.

        Args:
            other (PauliString): The right multiplicand.
            inplace (bool): If `True`, `self` is updated to hold the product.

        Returns:
            PauliString: The resultant circuit from right multiplying `self` with `other`.

        Raises:
            ValueError: If the lengths of the Pauli strings being multiplied differ.
        """
        if self._qubit_count != other._qubit_count:
            raise ValueError(
                f"Input Pauli string must be of length ({self._qubit_count}), "
                f"not {other._qubit_count}"
            )
        pauli_result = ""
        phase_result = self._phase * other._phase
        for i in range(self._qubit_count):
            # Are either identity?
            if i not in self._nontrivial and i not in other._nontrivial:
                pauli_result += "I"
            elif i not in self._nontrivial:
                pauli_result += other._nontrivial[i]
            elif i not in other._nontrivial:
                pauli_result += self._nontrivial[i]
            elif self._nontrivial[i] == other._nontrivial[i]:
                pauli_result += "I"
            else:
                gate, phase = _PRODUCT_MAP[self._nontrivial[i]][other._nontrivial[i]]
                pauli_result += gate
                phase_result *= phase

        # ignore complex global phase
        if phase_result.real < 0 or phase_result.imag < 0:
            pauli_result = "-" + pauli_result
        out_pauli_string = PauliString(pauli_result)

        if inplace:
            self._phase = out_pauli_string._phase
            self._qubit_count = out_pauli_string._qubit_count
            self._nontrivial = out_pauli_string._nontrivial
        return out_pauli_string

    def __mul__(self, other: PauliString) -> PauliString:
        """Right multiplication operator overload using `dot()`.

        Returns the result of multiplying the current circuit by the argument on its right.

        Args:
            other (PauliString): The right multiplicand.

        Returns:
            PauliString: The resultant circuit from right multiplying `self` with `other`.

        Raises:
            ValueError: If the lengths of the Pauli strings being multiplied differ.

        See Also:
            `braket.quantum_information.PauliString.dot()`
        """
        return self.dot(other)

    def __imul__(self, other: PauliString) -> PauliString:
        """Operator overload for right-multiplication assignment (`*=`) using `dot()`.

        Right-multiplies `self` by `other`, and assigns the result to `self`.

        Args:
            other (PauliString): The right multiplicand.

        Returns:
            PauliString: The resultant circuit from right multiplying `self` with `other`.

        Raises:
            ValueError: If the lengths of the Pauli strings being multiplied differ.

        See Also:
            `braket.quantum_information.PauliString.dot()`
        """
        return self.dot(other, inplace=True)

    def power(self, n: int, inplace: bool = False) -> PauliString:
        """Composes Pauli string with itself n times.

        Args:
            n (int): The number of times to self-multiply. Can be any integer value.
            inplace (bool): Update `self` if `True`

        Returns:
            PauliString: If `n` is positive, result from self-multiplication `n` times.
            If zero, identity. If negative, self-multiplication from trivial
            inverse (recall Pauli operators are involutory).

        Raises:
            ValueError: If `n` isn't a plain Python `int`.
        """
        if not isinstance(n, int):
            raise ValueError("Must be raised to integer power")

        # Since pauli ops involutory, result is either identity or unchanged
        pauli_other = PauliString(self)
        if n % 2 == 0:
            pauli_other._phase = 1
            pauli_other._nontrivial = {}

        if inplace:
            self._phase = pauli_other._phase
            self._qubit_count = pauli_other._qubit_count
            self._nontrivial = pauli_other._nontrivial
        return pauli_other

    def __pow__(self, n: int) -> PauliString:
        """Pow operator overload for Pauli string composition.

        Syntactic sugar for `power()`.

        Args:
            n (int): The number of times to self-multiply. Can be any integer

        Returns:
            PauliString: If `n` is positive, result from self-multiplication `n` times.
            If zero, identity. If negative, self-multiplication from trivial
            inverse (recall Pauli operators are involutory).

        Raises:
            ValueError: If `n` isn't a plain Python `int`.

        See Also:
            `braket.quantum_information.PauliString.power()`
        """
        return self.power(n)

    def __ipow__(self, n: int) -> PauliString:
        """Operator overload for in-place pow assignment (`**=`) using `power()`.

        Syntactic sugar for in-place `power()`.

        Args:
            n (int): The number of times to self-multiply. Can be any integer

        Returns:
            PauliString: If `n` is positive, result from self-multiplication `n` times.
            If zero, identity. If negative, self-multiplication from trivial
            inverse (recall Pauli operators are involutory).

        Raises:
            ValueError: If `n` isn't a plain Python `int`.

        See Also:
            `braket.quantum_information.PauliString.power()`
        """
        return self.power(n, inplace=True)

    def to_circuit(self) -> Circuit:
        """Returns circuit represented by this `PauliString`.

        Returns:
            Circuit: The circuit for this `PauliString`.
        """
        circ = Circuit()
        for qubit in range(self._qubit_count):
            if qubit not in self._nontrivial:
                circ = circ.i(qubit)
            elif self._nontrivial[qubit] == "X":
                circ = circ.x(qubit)
            elif self._nontrivial[qubit] == "Y":
                circ = circ.y(qubit)
            else:
                circ = circ.z(qubit)
        return circ

    def __eq__(self, other: PauliString):
        if isinstance(other, PauliString):
            return (
                self._phase == other._phase
                and self._nontrivial == other._nontrivial
                and self._qubit_count == other._qubit_count
            )
        return False

    def __getitem__(self, item: int):
        if item >= self._qubit_count:
            raise IndexError(item)
        return _PAULI_INDICES[self._nontrivial.get(item, "I")]

    def __len__(self):
        return self._qubit_count

    def __repr__(self):
        factors = [
            self._nontrivial[qubit] if qubit in self._nontrivial else "I"
            for qubit in range(self._qubit_count)
        ]
        return f"{PauliString._phase_to_str(self._phase)}{''.join(factors)}"

    @staticmethod
    def _split(pauli_word: str) -> tuple[int, str]:
        index = 0
        phase = 1
        if pauli_word[index] in {"+", "-"}:
            phase *= int(f"{pauli_word[index]}1")
            index += 1
        unsigned = pauli_word[index:]
        if not unsigned:
            raise ValueError("Pauli string cannot be empty")
        if set(unsigned) - _PAULI_INDICES.keys():
            raise ValueError(f"{pauli_word} is not a valid Pauli string")
        return phase, unsigned

    @staticmethod
    def _phase_to_str(phase: int) -> str:
        return "+" if phase > 0 else "-"

    def _generate_eigenstate_circuit(self, signs: tuple[int, ...]) -> Circuit:
        circ = Circuit()
        for qubit in range(len(signs)):
            state = signs[qubit] * self[qubit]
            if state == -3:
                circ.x(qubit)
            elif state == 1:
                circ.h(qubit)
            elif state == -1:
                circ.x(qubit).h(qubit)
            elif state == 2:
                circ.h(qubit).s(qubit)
            elif state == -2:
                circ.h(qubit).si(qubit)
        return circ



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json

import pytest

TEST_S3_OBJ_CONTENTS = {
    "TaskMetadata": {
        "Id": "blah",
        "Status": "COMPLETED",
    }
}


@pytest.fixture()
def s3_key(s3_resource, s3_bucket, s3_prefix):
    obj = s3_resource.Object(s3_bucket, f"{s3_prefix}/test_task_reading.json")

    try:
        obj_body = obj.get()["Body"].read().decode("utf-8")
        assert obj_body == json.dumps(TEST_S3_OBJ_CONTENTS)
    except s3_resource.meta.client.exceptions.NoSuchKey:
        # Put s3 object
        obj.put(ACL="private", Body=json.dumps(TEST_S3_OBJ_CONTENTS, indent=4))
    except AssertionError:
        # Put s3 object
        obj.put(ACL="private", Body=json.dumps(TEST_S3_OBJ_CONTENTS, indent=4))

    return obj.key


def test_retrieve_s3_object_body(aws_session, s3_bucket, s3_key):
    obj_body = aws_session.retrieve_s3_object_body(s3_bucket, s3_key)
    assert obj_body == json.dumps(TEST_S3_OBJ_CONTENTS, indent=4)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from decimal import Decimal

import pytest

from braket.timings.time_series import StitchBoundaryCondition, TimeSeries, _all_close


@pytest.fixture
def default_values():
    return [(2700, 25.1327), (300, 25.1327), (600, 15.1327), (Decimal(0.3), Decimal(0.4))]


@pytest.fixture
def default_time_series(default_values):
    time_series = TimeSeries()
    for value in default_values:
        time_series.put(value[0], value[1])
    return time_series


def test_add_chaining():
    time_series = (
        TimeSeries()
        .put(time=0, value=0)
        .put(300, 25.1327)
        .put(2700, 25.1327)
        .put(3000, 0)
        .put(2700, 25.1327)
    )
    assert len(time_series) == 4


def test_iteration_sorted(default_values, default_time_series):
    sorted_returned_values = [(item.time, item.value) for item in default_time_series]
    assert sorted_returned_values == sorted(default_values)


def test_get_sorted(default_values, default_time_series):
    sorted_values = sorted(default_values)
    assert default_time_series.times() == [item[0] for item in sorted_values]
    assert default_time_series.values() == [item[1] for item in sorted_values]


def test_constant_like():
    times = list(range(10))
    constant_ts = TimeSeries.constant_like(times, constant=3.14)
    assert times == constant_ts.times()
    assert constant_ts.values() == [3.14] * len(times)


def test_periodic_signal():
    times = list(range(4))
    values = [0, 1, 3, 0]
    new_ts = TimeSeries.periodic_signal(times=times, values=values, num_repeat=3)
    expected_times = list(range(10))
    expected_values = [0, 1, 3, 0, 1, 3, 0, 1, 3, 0]

    assert new_ts.times() == expected_times
    assert new_ts.values() == expected_values


def test_constant_like_with_time_series():
    time_series = TimeSeries().put(0.0, 0.0).put(1.2, 3.14)
    constant_ts = TimeSeries.constant_like(time_series, constant=3.14)
    times = time_series.times()
    assert times == constant_ts.times()
    assert constant_ts.values() == [3.14] * len(times)


@pytest.mark.parametrize(
    "area, value_max, slew_rate_max, time_separation_min",
    [
        (2.0, 2.0, 4.0, 1.0),
        (1.0, 2.0, 4.0, 1.0),
        (4.0, 2.0, 1.0, 1.0),
        (1.0, 2.0, 1.0, 1.0),
    ],
)
def test_trapezoidal_signal_as_triangular_signal(
    area, value_max, slew_rate_max, time_separation_min
):
    ts = TimeSeries.trapezoidal_signal(area, value_max, slew_rate_max, time_separation_min)
    t_ramp = max(time_separation_min, value_max / slew_rate_max)
    value = area / t_ramp
    assert ts.times() == [0, t_ramp, 2 * t_ramp]
    assert ts.values() == [0, value, 0]


@pytest.mark.parametrize(
    "area, value_max, slew_rate_max, time_separation_min",
    [
        (4.0, 2.0, 4.0, 1.0),
        (2.1, 2.0, 4.0, 1.0),
        (6.0, 2.0, 1.0, 1.0),
        (4.1, 2.0, 1.0, 1.0),
    ],
)
def test_trapezoidal_signal_as_min_trapezoidal_signal(
    area, value_max, slew_rate_max, time_separation_min
):
    ts = TimeSeries.trapezoidal_signal(area, value_max, slew_rate_max, time_separation_min)
    t_ramp = max(time_separation_min, value_max / slew_rate_max)
    value = area / (t_ramp + time_separation_min)
    assert ts.times() == [0, t_ramp, t_ramp + time_separation_min, 2 * t_ramp + time_separation_min]
    assert ts.values() == [0, value, value, 0]


@pytest.mark.parametrize(
    "area, value_max, slew_rate_max, time_separation_min",
    [
        (4.1, 2.0, 4.0, 1.0),
        (6.1, 2.0, 1.0, 1.0),
    ],
)
def test_trapezoidal_signal_with_large_area(area, value_max, slew_rate_max, time_separation_min):
    ts = TimeSeries.trapezoidal_signal(area, value_max, slew_rate_max, time_separation_min)
    t_ramp = max(time_separation_min, value_max / slew_rate_max)
    t_plateau = area / value_max - t_ramp
    assert ts.times() == [0, t_ramp, t_ramp + t_plateau, 2 * t_ramp + t_plateau]
    assert ts.values() == [0, value_max, value_max, 0]


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "area, value_max, slew_rate_max, time_separation_min",
    [
        (-4.1, 2.0, 4.0, 1.0),
        (4.1, -2.0, 4.0, 1.0),
        (4.1, 2.0, -4.0, 1.0),
        (4.1, 2.0, 4.0, -1.0),
    ],
)
def test_trapezoidal_signal_with_negative_para(area, value_max, slew_rate_max, time_separation_min):
    TimeSeries.trapezoidal_signal(area, value_max, slew_rate_max, time_separation_min)


@pytest.mark.xfail(raises=ValueError)
def test_periodic_signal_not_eq_length():
    times = list(range(5))
    values = [0.5, 1, 1, 0]
    TimeSeries.periodic_signal(times=times, values=values, num_repeat=3)


def test_concatenate():
    times_1 = list(range(4))
    values_1 = [0.5, 1, 1, 0]
    time_series_1 = TimeSeries.from_lists(times=times_1, values=values_1)

    times_2 = list(range(4, 8))
    values_2 = [-0.5, -1, -1, 0]
    time_series_2 = TimeSeries.from_lists(times=times_2, values=values_2)

    new_ts = time_series_1.concatenate(time_series_2)

    assert new_ts.times() == times_1 + times_2
    assert new_ts.values() == values_1 + values_2

    new_ts = time_series_1.concatenate(TimeSeries())
    assert new_ts.times() == times_1
    assert new_ts.values() == values_1

    new_ts = TimeSeries().concatenate(time_series_1)
    assert new_ts.times() == times_1
    assert new_ts.values() == values_1

    new_ts = TimeSeries().concatenate(TimeSeries())
    assert new_ts.times() == []
    assert new_ts.values() == []


@pytest.mark.xfail(raises=ValueError)
def test_concatenate_not_ordered():
    times_1 = list(range(4))
    values_1 = [0.5, 1, 1, 0]
    time_series_1 = TimeSeries.from_lists(times=times_1, values=values_1)

    times_2 = list(range(4))
    values_2 = [-0.5, -1, -1, 0]
    time_series_2 = TimeSeries.from_lists(times=times_2, values=values_2)

    time_series_1.concatenate(time_series_2)


def test_from_lists():
    times = list(range(4))
    values = [0.5, 1, 1, 0]
    ts = TimeSeries.from_lists(times=times, values=values)
    assert ts.times() == times
    assert ts.values() == values


@pytest.mark.xfail(raises=ValueError)
def test_from_lists_not_equal_size():
    times = list(range(4))
    values = [0.5, 1, 1]
    TimeSeries.from_lists(times=times, values=values)


def test_stitch():
    times_1 = list(range(4))
    values_1 = [0.5, 1, 1, 0]
    time_series_1 = TimeSeries.from_lists(times=times_1, values=values_1)

    times_2 = list(range(4))
    values_2 = [-0.5, -1, -1, 0]
    time_series_2 = TimeSeries.from_lists(times=times_2, values=values_2)

    new_ts_mean = time_series_1.stitch(time_series_2, boundary=StitchBoundaryCondition.MEAN)
    new_ts_left = time_series_1.stitch(time_series_2, boundary=StitchBoundaryCondition.LEFT)
    new_ts_right = time_series_1.stitch(time_series_2, boundary=StitchBoundaryCondition.RIGHT)

    excepted_times = list(range(7))
    assert new_ts_mean.times() == excepted_times
    assert new_ts_left.times() == excepted_times
    assert new_ts_right.times() == excepted_times

    assert new_ts_mean.values() == [0.5, 1, 1, -0.25, -1, -1, 0]
    assert new_ts_left.values() == [0.5, 1, 1, 0, -1, -1, 0]
    assert new_ts_right.values() == [0.5, 1, 1, -0.5, -1, -1, 0]


def test_stitch_empty_ts():
    times = list(range(4))
    values = [0.5, 1, 1, 0]
    time_series = TimeSeries.from_lists(times=times, values=values)
    new_ts = time_series.stitch(TimeSeries())
    assert new_ts.times() == times
    assert new_ts.values() == values

    new_ts = TimeSeries().stitch(TimeSeries())
    assert new_ts.times() == []
    assert new_ts.values() == []

    new_ts = TimeSeries().stitch(time_series)
    assert new_ts.times() == time_series.times()
    assert new_ts.values() == time_series.values()


@pytest.mark.xfail(raises=ValueError)
def test_stitch_wrong_bndry_value():
    times_1 = list(range(4))
    values_1 = [0.5, 1, 1, 0]
    time_series_1 = TimeSeries.from_lists(times=times_1, values=values_1)

    times_2 = list(range(4))
    values_2 = [-0.5, -1, -1, 0]
    time_series_2 = TimeSeries.from_lists(times=times_2, values=values_2)

    time_series_1.stitch(time_series_2, boundary="average")


@pytest.mark.parametrize(
    "time_res, expected_times",
    [
        # default_time_series: [(Decimal(0.3), Decimal(0.4), (300, 25.1327), (600, 15.1327), (2700, 25.1327))] # noqa
        (Decimal(0.5), [Decimal("0.5"), Decimal("300"), Decimal("600"), Decimal("2700")]),
        (Decimal(1), [Decimal("0"), Decimal("300"), Decimal("600"), Decimal("2700")]),
        (Decimal(200), [Decimal("0"), Decimal("400"), Decimal("600"), Decimal("2800")]),
        (Decimal(1000), [Decimal("0"), Decimal("1000"), Decimal("3000")]),
    ],
)
def test_discretize_times(default_time_series, time_res, expected_times):
    value_res = Decimal("1")
    assert expected_times == default_time_series.discretize(time_res, value_res).times()


@pytest.mark.parametrize(
    "value_res, expected_values",
    [
        # default_time_series: [(Decimal(0.3), Decimal(0.4), (300, 25.1327), (600, 15.1327), (2700, 25.1327))] # noqa
        (Decimal("0.1"), [Decimal("0.4"), Decimal("25.1"), Decimal("15.1"), Decimal("25.1")]),
        (Decimal(1), [Decimal("0"), Decimal("25"), Decimal("15"), Decimal("25")]),
        (Decimal(6), [Decimal("0"), Decimal("24"), Decimal("18"), Decimal("24")]),
        (Decimal(100), [Decimal("0"), Decimal("0"), Decimal("0"), Decimal("0")]),
    ],
)
def test_discretize_values(default_time_series, value_res, expected_values):
    time_res = Decimal("0.1")
    assert expected_values == default_time_series.discretize(time_res, value_res).values()


@pytest.mark.parametrize(
    "first_series, second_series, expected_result",
    [
        (TimeSeries(), TimeSeries(), True),
        (TimeSeries().put(0.1, 0.2), TimeSeries(), False),
        (TimeSeries().put(float(0.1), float(0.2)), TimeSeries().put(float(0.1), float(0.2)), True),
        (TimeSeries().put(float(1), float(0.2)), TimeSeries().put(int(1), float(0.2)), True),
        (TimeSeries().put(float(0.1), float(0.2)), TimeSeries().put(float(0.2), float(0.2)), False),
        (TimeSeries().put(float(0.1), float(0.3)), TimeSeries().put(float(0.1), float(0.2)), False),
    ],
)
def test_all_close(first_series, second_series, expected_result):
    result = _all_close(first_series, second_series)
    assert result == expected_result



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest
from oqpy import PortVar
from oqpy.base import expr_matches

from braket.pulse import Port


@pytest.fixture
def port_id():
    return "test_port_ff"


@pytest.fixture
def port_time_resolution():
    return 1e-9


@pytest.fixture
def port_properties():
    return {"dt": 1e-6, "direction": "tx"}


@pytest.fixture
def port(port_id, port_time_resolution, port_properties):
    return Port(port_id, port_time_resolution, port_properties)


def test_port_no_properties(port_id, port_time_resolution):
    port = Port(port_id, port_time_resolution)
    assert port.id == port_id
    assert port.dt == port_time_resolution
    assert port.properties is None


def test_port_to_oqpy_expression(port, port_id):
    expected_expression = PortVar(port_id)
    assert expr_matches(port._to_oqpy_expression(), expected_expression)


def test_port_equality(port, port_time_resolution):
    p2 = Port(port.id, port_time_resolution)
    p3 = Port("random_id", port_time_resolution, properties=port.properties)
    p4 = Port(port.id, port_time_resolution, properties={"random_property": "foo"})
    p5 = Port(port.id, 0)
    assert port == p2
    assert port == p4
    assert port == p5
    assert port != p3



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import warnings
from abc import ABC, abstractmethod
from typing import Any, Optional, Union

from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.annealing.problem import Problem
from braket.circuits import Circuit, Noise
from braket.circuits.noise_model import NoiseModel
from braket.circuits.translations import SUPPORTED_NOISE_PRAGMA_TO_NOISE
from braket.device_schema import DeviceActionType
from braket.ir.openqasm import Program
from braket.tasks.quantum_task import QuantumTask
from braket.tasks.quantum_task_batch import QuantumTaskBatch


class Device(ABC):
    """An abstraction over quantum devices that includes quantum computers and simulators."""

    def __init__(self, name: str, status: str):
        """Initializes a `Device`.

        Args:
            name (str): Name of quantum device
            status (str): Status of quantum device
        """
        self._name = name
        self._status = status

    @abstractmethod
    def run(
        self,
        task_specification: Union[Circuit, Problem],
        shots: Optional[int],
        inputs: Optional[dict[str, float]],
        *args,
        **kwargs,
    ) -> QuantumTask:
        """Run a quantum task specification on this quantum device. A quantum task can be a circuit
        or an annealing problem.

        Args:
            task_specification (Union[Circuit, Problem]): Specification of a quantum task
                to run on device.
            shots (Optional[int]): The number of times to run the quantum task on the device.
                Default is `None`.
            inputs (Optional[dict[str, float]]): Inputs to be passed along with the
                IR. If IR is an OpenQASM Program, the inputs will be updated with this value.
                Not all devices and IR formats support inputs. Default: {}.
            *args (Any):  Arbitrary arguments.
            **kwargs (Any): Arbitrary keyword arguments.

        Returns:
            QuantumTask: The QuantumTask tracking task execution on this device
        """

    @abstractmethod
    def run_batch(
        self,
        task_specifications: Union[
            Union[Circuit, Problem],
            list[Union[Circuit, Problem]],
        ],
        shots: Optional[int],
        max_parallel: Optional[int],
        inputs: Optional[Union[dict[str, float], list[dict[str, float]]]],
        *args: Any,
        **kwargs: Any,
    ) -> QuantumTaskBatch:
        """Executes a batch of quantum tasks in parallel

        Args:
            task_specifications (Union[Union[Circuit, Problem], list[Union[Circuit, Problem]]]):
                Single instance or list of circuits or problems to run on device.
            shots (Optional[int]): The number of times to run the circuit or annealing problem.
            max_parallel (Optional[int]): The maximum number of quantum tasks to run  in parallel.
                Batch creation will fail if this value is greater than the maximum allowed
                concurrent quantum tasks on the device.
            inputs (Optional[Union[dict[str, float], list[dict[str, float]]]]): Inputs to be
                passed along with the IR. If the IR supports inputs, the inputs will be updated
                with this value.
            *args (Any):  Arbitrary arguments.
            **kwargs (Any): Arbitrary keyword arguments.

        Returns:
            QuantumTaskBatch: A batch containing all of the qauntum tasks run
        """

    @property
    def name(self) -> str:
        """Return the name of this Device.

        Returns:
            str: The name of this Device
        """
        return self._name

    @property
    def status(self) -> str:
        """Return the status of this Device.

        Returns:
            str: The status of this Device
        """
        return self._status

    def _validate_device_noise_model_support(self, noise_model: NoiseModel) -> None:
        supported_noises = set(
            SUPPORTED_NOISE_PRAGMA_TO_NOISE[pragma].__name__
            for pragma in self.properties.action[DeviceActionType.OPENQASM].supportedPragmas
            if pragma in SUPPORTED_NOISE_PRAGMA_TO_NOISE
        )
        noise_operators = set(noise_instr.noise.name for noise_instr in noise_model._instructions)
        if not noise_operators <= supported_noises:
            raise ValueError(
                f"{self.name} does not support noise simulation or the noise model includes noise "
                + f"that is not supported by {self.name}."
            )

    def _apply_noise_model_to_circuit(
        self, task_specification: Union[Circuit, Problem, Program, AnalogHamiltonianSimulation]
    ) -> None:
        if isinstance(task_specification, Circuit):
            for instruction in task_specification.instructions:
                if isinstance(instruction.operator, Noise):
                    warnings.warn(
                        "The noise model of the device is applied to a circuit that already has"
                        " noise instructions."
                    )
                    break
            task_specification = self._noise_model.apply(task_specification)
        else:
            warnings.warn(
                "Noise model is only applicable to circuits. The type of the task specification is"
                f" {task_specification.__class__.__name__}. The noise model of the device does not"
                " apply."
            )
        return task_specification
